import os
import sys
sys.path.append('')
from dotenv import load_dotenv
import pandas as pd
import json
from hopp.simulation.technologies.sites import SiteInfo
from hopp.simulation.technologies.sites import flatirons_site as sample_site
from hopp.utilities.keys import set_developer_nrel_gov_key
# from plot_reopt_results import plot_reopt_results
# from run_reopt import run_reopt
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import hopp_for_h2
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import run_h2a as run_h2a #no h2a function
from greenheart.to_organize.H2_Analysis.simple_dispatch import SimpleDispatch
from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals
import hopp.simulation.technologies.hydrogen.electrolysis.run_h2_PEM as run_h2_PEM
import numpy as np
import numpy_financial as npf
from lcoe.lcoe import lcoe as lcoe_calc
import matplotlib.pyplot as plt
import warnings
from pathlib import Path
warnings.filterwarnings("ignore")
from hopp.tools.resource import *
from hopp.tools.resource.resource_loader import site_details_creator

from greenheart.to_organize import hopp_tools_steel
import copy
from greenheart.to_organize import plot_results
from greenheart.to_organize import run_profast_for_hydrogen
from greenheart.to_organize.hopp_tools_steel import hoppDict
import yaml

"""
OSW-H2 LCOH analysis
"""

#Set API key
load_dotenv()
NREL_API_KEY = os.getenv("NREL_API_KEY")
set_developer_nrel_gov_key('NREL_API_KEY')  # Set this key manually here if you are not setting it using the .env

#Step 1: User Inputs for scenario
save_hybrid_plant_yaml = True # hybrid_plant requires special processing of the SAM objects
save_model_input_yaml = True # saves the inputs for each model/major function
save_model_output_yaml = True # saves the outputs for each model/major function

resource_year = 2013
atb_years = [
            2025,
            # 2030,
            # 2035
            ]
'''
Tech neutral ITC and PTC for years after 2022
Wind PTC [$/kWh] is in 1992 dollars adjusted for inflation in run_pyfast_for_hydrogen.py
H2 PTC [$/kg] is in 2022 dollars adjusted for inflation in run_pyfast_for_hydrogen.py
Wind ITC [%] applied to wind capital expenditure and HVDC cabling in run_pyfast_for_hydrogen.py
Base: Base credit only 100% full valuation. No prevailing wages. No bonus credits.
Max: 100% full valuation credit. Prevailing wages met.
Bonus: 100% full valuation credit. Prevailing wages met. Bonus content credit met.
'''
policy = {
    'No Policy': {'Wind ITC': 0, 'Wind PTC': 0, "H2 PTC": 0},
    # 'Base PTC': {'Wind ITC': 0, 'Wind PTC': 0.003, "H2 PTC": 0.60},
    # 'Max PTC': {'Wind ITC': 0, 'Wind PTC': 0.015, "H2 PTC": 3.00},
    # 'Base ITC': {'Wind ITC': 0.06, 'Wind PTC': 0.00, "H2 PTC": 0.60},
    # 'Max ITC': {'Wind ITC': 0.30, 'Wind PTC': 0.00, "H2 PTC": 3.00},
    # 'Bonus PTC': {'Wind ITC': 0, 'Wind PTC': 0.0165, "H2 PTC": 3.00},
    # 'Bonus ITC': {'Wind ITC': 0.40, 'Wind PTC': 0.00, "H2 PTC": 3.00}
}

sample_site['year'] = resource_year
useful_life = 30
critical_load_factor = 1
run_reopt_flag = False
custom_powercurve = False   #A flag that is applicable when using PySam WindPower (not FLORIS)
storage_used = False
battery_can_grid_charge = False
grid_connected_hopp = False
floris = True      #TODO: Set to True and get the floris files working

# Technology sizing
interconnection_size_mw = 1000
electrolyzer_size_mw = 1000
wind_size_mw = 1000
solar_size_mw = 0
storage_size_mw = 0
storage_size_mwh = 0

#TODO: Should all turbines be used for all years?
turbine_name = [
                '12MW',
                # '15MW',
                # '18MW'
                ]


scenario_choice = 'Offshore Wind-H2 Analysis'

site_selection = [
                'Site 1',
                # 'Site 2',
                # 'Site 3',
                # 'Site 4'
                ]

scenario = dict()
kw_continuous = electrolyzer_size_mw * 1000
load = [kw_continuous for x in
        range(0, 8760)]  # * (sin(x) + pi) Set desired/required load profile for plant

#Site lat and lon will be set by data loaded from Orbit runs

# Financial inputs
discount_rate = 0.10 # nominal return based on 2022 ATB basline workbook
debt_equity_split = 68.5    # 2022 ATB uses 68.5% debt

# Wind costs input from ORBIT analysis
h2_model ='Simple'  #Basic cost model based on H2a and HFTO program record for PEM electrolysis
# h2_model = 'H2A'

# These inputs are not used in this analysis (no solar or storage)
solar_cost_kw = 9999
storage_cost_kw = 250
storage_cost_kwh = 240

# Flags (TODO: remove them and update documentation)
forced_sizes = True # no REopt
force_electrolyzer_cost = False


# Enable Ability to purchase/sell electricity to/from grid. Price Defined in $/kWh
# sell_price = 0.01
# buy_price = 0.01
sell_price = False
buy_price = False

# Set paths for results, floris and orbit
parent_path = os.path.abspath('')
results_dir = parent_path + '/examples/H2_Analysis/results/'
results = pd.DataFrame()
floris_dir = parent_path + '/floris_input_files/'

# Site specific turbine information
path = ('examples/H2_Analysis/OSW_H2_sites_turbines_and_costs.xlsx')
xl = pd.ExcelFile(path)

# which plots to show
plot_power_production = True
plot_battery = True
plot_grid = True
plot_h2 = True
plot_desal = True
plot_wind = True
plot_hvdcpipe = True
plot_hvdcpipe_lcoh = True


for option in policy:
    for atb_year in atb_years:
        for site_location in site_selection:
            site_number = site_location.split(' ')[1]
            for turbine_model in turbine_name:

                hopp_dict = hoppDict(save_model_input_yaml, save_model_output_yaml)
                sub_dict = {
                    'policy': policy[option],
                    'atb_year': atb_year,
                    'site_location': site_location,
                    'parent_path': parent_path,
                    'load': load,
                    'kw_continuous': kw_continuous,
                    'sample_site': sample_site,
                    'discount_rate': discount_rate,
                    'forced_sizes': forced_sizes,
                    'force_electrolyzer_cost': force_electrolyzer_cost,
                    'wind_size': wind_size_mw,
                    'solar_size': solar_size_mw,
                    'storage_size_mw': storage_size_mw,
                    'storage_size_mwh': storage_size_mwh,
                    'solar_cost_kw': solar_cost_kw,
                    # 'storage_cost_kw': storage_cost_kw,
                    # 'storage_cost_kwh': storage_cost_kwh,
                    'debt_equity_split': debt_equity_split,
                    'useful_life': useful_life,
                    'critical_load_factor': critical_load_factor,
                    'run_reopt_flag': run_reopt_flag,
                    'custom_powercurve': custom_powercurve,
                    'storage_used': storage_used,
                    'battery_can_grid_charge': battery_can_grid_charge,
                    'grid_connected_hopp': grid_connected_hopp,
                    'interconnection_size_mw': interconnection_size_mw,
                    'electrolyzer_size_mw': electrolyzer_size_mw,
                    'scenario':
                        {
                            'Useful Life': useful_life,
                            'Debt Equity': debt_equity_split,
                            'discount_rate': discount_rate,
                        },
                    'sell_price': False,
                    'buy_price': False,
                    'h2_model': h2_model,
                    'results_dir': results_dir,
                    'scenario_choice': scenario_choice,
                }

                hopp_dict.add('Configuration', sub_dict)

                plot_dict = {
                    'plot':
                        {
                            'plot_power_production': True,
                            'plot_battery': True,
                            'plot_grid': True,
                            'plot_h2': True,
                            'plot_desal': True,
                            'plot_wind': True,
                            'plot_hvdcpipe': True,
                            'plot_hvdcpipe_lcoh': True,
                        }
                }

                hopp_dict.add('Configuration', plot_dict)


                # set policy values
                hopp_dict, scenario, policy_option = hopp_tools_steel.set_policy_values(hopp_dict, scenario, policy, option)
                print(scenario['Wind PTC'])

                turbinesheet = turbine_model[-4:]
                scenario_df = xl.parse(turbinesheet)
                scenario_df.set_index(["Parameter"], inplace = True)

                site_df = scenario_df[site_location]

                # turbine_model = str(site_df['Turbine rating'])+'MW'

                turbine_rating = site_df['Turbine rating']

                # set turbine values
                hopp_dict, scenario, nTurbs, floris_config = hopp_tools_steel.set_turbine_model(hopp_dict, turbine_model, scenario, parent_path,floris_dir, floris)

                scenario['Useful Life'] = useful_life

                # financials
                hopp_dict, scenario = hopp_tools_steel.set_financial_info(hopp_dict, scenario, debt_equity_split, discount_rate)

                # set electrolyzer information
                #TODO: Verify that we want to use conservative replacement scenario and costs for different years
                electrolysis_scale = 'Centralized'
                electrolyzer_replacement_scenario = 'Conservative'
                hopp_dict, electrolyzer_capex_kw,capex_ratio_dist, electrolyzer_energy_kWh_per_kg, time_between_replacement =  hopp_tools_steel.set_electrolyzer_info(hopp_dict,atb_year,electrolysis_scale,electrolyzer_replacement_scenario,turbine_rating,direct_coupling = False)

                # Extract Scenario Information from ORBIT Runs
                # Load Excel file of scenarios
                # OSW sites and cost file including turbines 8/16/2022

                # site info
                hopp_dict, site_df, sample_site = hopp_tools_steel.set_site_info(hopp_dict, site_df, sample_site)

                site_name = site_df['Representative region']
                fixed_or_floating_wind = site_df['Substructure technology']
                site = SiteInfo(sample_site, hub_height=scenario['Tower Height'])

                hopp_dict.add('Configuration', {'site': site})

                #Assign Orbit results to scenario cost details
                total_capex = site_df['Total CapEx']
                wind_cost_kw = copy.deepcopy(total_capex)
                wind_om_cost_kw = site_df['OpEx, $/kW-yr']
                wind_net_cf = site_df['Assumed NCF']

                # set export financials
                wind_cost_kw, wind_om_cost_kw, total_export_system_cost, total_export_om_cost = hopp_tools_steel.set_export_financials(wind_size_mw,
                                                                                                                                wind_cost_kw,
                                                                                                                                wind_om_cost_kw,
                                                                                                                                useful_life,
                                                                                                                                site_df)

                # set wind financials
                new_wind_cost_kw, new_wind_om_cost_kw, new_wind_net_cf = hopp_tools_steel.set_turbine_financials(turbine_model,
                                                                                                            fixed_or_floating_wind,
                                                                                                            atb_year,
                                                                                                            wind_cost_kw,
                                                                                                            wind_om_cost_kw,
                                                                                                            wind_net_cf,
                                                                                                            parent_path)

                wind_cost_kw = copy.deepcopy(new_wind_cost_kw)
                wind_om_cost_kw = copy.deepcopy(new_wind_om_cost_kw)


                #Plot Wind Data to ensure offshore data is sound
                wind_data = site.wind_resource._data['data']
                wind_speed = [W[2] for W in wind_data]
                plot_results.plot_wind_results(wind_data, site_name, site_df['Representative coordinates'], results_dir, plot_wind)
                print("Wind speed check: ", np.max(wind_speed))

                # Run HOPP
                print('Site name: ', site_name, 'Turbine rating: ',turbine_rating, 'Total Capex: ', total_capex, 'ATB Capex', new_wind_cost_kw)
                hopp_dict, combined_hybrid_power_production_hopp, energy_shortfall_hopp, combined_hybrid_curtailment_hopp, hybrid_plant, wind_size_mw, solar_size_mw, lcoe = \
                hopp_tools_steel.run_HOPP(
                    hopp_dict,
                    scenario,
                    site,
                    sample_site,
                    forced_sizes,
                    solar_size_mw,
                    wind_size_mw,
                    storage_size_mw,
                    storage_size_mwh,
                    wind_cost_kw,
                    solar_cost_kw,
                    storage_cost_kw,
                    storage_cost_kwh,
                    kw_continuous,
                    load,
                    electrolyzer_size_mw,
                    wind_om_cost_kw,
                    nTurbs,
                    floris_config,
                    floris,
                )

                wind_plant_size = hybrid_plant.wind.system_capacity_kw
                print('Wind plant size: ',hybrid_plant.wind.system_capacity_kw,\
                    )

                #Step 4: Plot HOPP Results
                plot_results.plot_HOPP(combined_hybrid_power_production_hopp,
                                        energy_shortfall_hopp,
                                        combined_hybrid_curtailment_hopp,
                                        load,
                                        results_dir,
                                        site_name,
                                        atb_year,
                                        turbine_model,
                                        hybrid_plant,
                                        plot_power_production)

                #Step 5: Run Simple Dispatch Model (no battery is used in system)
                hopp_dict, combined_pv_wind_storage_power_production_hopp, battery_SOC, battery_used, excess_energy = \
                    hopp_tools_steel.run_battery(
                    hopp_dict,
                    energy_shortfall_hopp,
                    combined_hybrid_curtailment_hopp,
                    combined_hybrid_power_production_hopp
                    )

                plot_results.plot_battery_results(
                    combined_hybrid_curtailment_hopp,
                    energy_shortfall_hopp,
                    combined_pv_wind_storage_power_production_hopp,
                    combined_hybrid_power_production_hopp,
                    battery_SOC,
                    battery_used,
                    results_dir,
                    site_name,atb_year,turbine_model,
                    load,
                    plot_battery,
                )

                # grid information (no grid is used in system)
                hopp_dict, cost_to_buy_from_grid, profit_from_selling_to_grid, energy_to_electrolyzer = hopp_tools_steel.grid(
                    hopp_dict,
                    combined_pv_wind_storage_power_production_hopp,
                    sell_price,
                    excess_energy,
                    buy_price,
                    kw_continuous,
                    plot_grid,
                )

                #Step 6: Run the H2_PEM model
                h2_model = 'Simple'
                hopp_dict, H2_Results, H2A_Results, electrical_generation_timeseries = hopp_tools_steel.run_H2_PEM_sim(
                    hopp_dict,
                    hybrid_plant,
                    energy_to_electrolyzer,
                    scenario,
                    wind_size_mw,
                    solar_size_mw,
                    electrolyzer_size_mw,
                    kw_continuous,
                    electrolyzer_capex_kw,
                    lcoe,
                )

                plot_results.plot_h2_results(H2_Results,
                                electrical_generation_timeseries,
                                results_dir,
                                site_name,atb_year,turbine_model,
                                load,
                                plot_h2)

                #Step 6b: Run desal model
                hopp_dict, desal_capex, desal_opex = hopp_tools_steel.desal_model(
                    hopp_dict,
                    H2_Results,
                    electrolyzer_size_mw,
                    electrical_generation_timeseries,
                    useful_life,
                )

                # compressor model
                hopp_dict,compressor, compressor_results = hopp_tools_steel.compressor_model(hopp_dict)

                #Pressure Vessel Model Example
                hopp_dict,storage_input, storage_output = hopp_tools_steel.pressure_vessel(hopp_dict)

                # pipeline model
                total_h2export_system_cost, opex_pipeline, dist_to_port_value = hopp_tools_steel.pipeline(site_df,
                                                                                H2_Results,
                                                                                useful_life,
                                                                                storage_input)


                # plot HVDC vs pipe
                plot_results.plot_hvdcpipe(total_export_system_cost,
                                            total_h2export_system_cost,
                                            site_name,
                                            atb_year,
                                            dist_to_port_value,
                                            results_dir)

                revised_renewable_cost = hybrid_plant.grid.total_installed_cost
                hydrogen_annual_production = H2_Results['hydrogen_annual_output']
                hydrogen_storage_capacity_kg = 0
                hydrogen_storage_cost_USDprkg = 0
                water_cost = 0.006868 #($/gal) average of green steel sites' water cost


                h2_ptc = scenario['H2 PTC']
                wind_ptc = scenario['Wind PTC']
                wind_itc = scenario['Wind ITC']

                print(revised_renewable_cost)
                #Run HVDC export scenario
                export_hvdc = True  #HVDC scenario
                h2a_solution,h2a_summary,lcoh_breakdown,electrolyzer_installed_cost_kw = run_profast_for_hydrogen.run_profast_for_hydrogen(site_name,electrolyzer_size_mw,H2_Results,\
                                electrolyzer_capex_kw,time_between_replacement,hydrogen_storage_capacity_kg,hydrogen_storage_cost_USDprkg,\
                                desal_capex,desal_opex,useful_life,atb_year,water_cost,wind_size_mw,solar_size_mw, \
                                hybrid_plant,revised_renewable_cost,wind_om_cost_kw,
                                total_export_system_cost,
                                total_export_om_cost,
                                export_hvdc, combined_pv_wind_storage_power_production_hopp,
                                grid_connected_hopp,h2_ptc,wind_ptc,wind_itc)
                lcoh = h2a_solution['price']
                print('LCOH: ', lcoh)
                # # Max hydrogen production rate [kg/hr]
                max_hydrogen_production_rate_kg_hr = np.max(H2_Results['hydrogen_hourly_production'])
                max_hydrogen_delivery_rate_kg_hr  = np.mean(H2_Results['hydrogen_hourly_production'])

                electrolyzer_capacity_factor = H2_Results['cap_factor']

                test = dict()

                test['Site Name'] = site_name
                test['ATB Year'] = atb_year
                test['HVDC Export'] = export_hvdc
                test['Pipeline Export'] = False
                test['Plant life'] = useful_life
                test['Policy'] = option
                test['LCOH: total ($/kg)'] = lcoh_breakdown['LCOH: total ($/kg)']
                test['Turbine size (MW)'] = turbine_rating
                test['Wind Plant size (MW)'] = wind_size_mw
                test['Wind Plant Size Adjusted for Turbine Rating(MW)'] = wind_plant_size /1000
                test['Electrolyzer size (MW)'] = electrolyzer_size_mw
                test['Load Profile (kW)'] = kw_continuous
                test['Energy to Electrolyzer (kW)'] = np.sum(energy_to_electrolyzer)
                test['Wind capacity factor (%)'] = hybrid_plant.wind.capacity_factor
                test['Electrolyzer capacity factor (%)'] = electrolyzer_capacity_factor
                test['H2 production (kg)'] = hydrogen_annual_production
                test['LCOH ($/kg)'] = lcoh
                test['LCOH: Compression & storage ($/kg)'] = lcoh_breakdown['LCOH: Compression & storage ($/kg)']
                test['LCOH: Electrolyzer CAPEX ($/kg)']= lcoh_breakdown['LCOH: Electrolyzer CAPEX ($/kg)']
                test['LCOH: Desalination CAPEX ($/kg)'] = lcoh_breakdown['LCOH: Desalination CAPEX ($/kg)']
                test['LCOH: HVDC Export CAPEX ($/kg)'] = lcoh_breakdown['LCOH: HVDC Export CAPEX ($/kg)']
                test['LCOH: Pipeline Export CAPEX ($/kg)'] = lcoh_breakdown['LCOH: Pipeline Export CAPEX ($/kg)']
                test['LCOH: HVDC Export FOM ($/kg)'] = lcoh_breakdown['LCOH: HVDC Export FOM ($/kg)']
                test['LCOH: Pipeline Export FOM ($/kg)'] = lcoh_breakdown['LCOH: Pipeline Export FOM ($/kg)']
                test['LCOH: Electrolyzer FOM ($/kg)'] = lcoh_breakdown['LCOH: Electrolyzer FOM ($/kg)']
                test['LCOH: Electrolyzer VOM ($/kg)']=lcoh_breakdown['LCOH: Electrolyzer VOM ($/kg)']
                test['LCOH: Desalination FOM ($/kg)'] = lcoh_breakdown['LCOH: Desalination FOM ($/kg)']
                test['LCOH: Renewable plant ($/kg)'] = lcoh_breakdown['LCOH: Renewable plant ($/kg)']
                test['LCOH: Renewable FOM ($/kg)']= lcoh_breakdown['LCOH: Renewable FOM ($/kg)']
                test['LCOH: Taxes ($/kg)']=lcoh_breakdown['LCOH: Taxes ($/kg)']
                test['LCOH: Water consumption ($/kg)'] = lcoh_breakdown['LCOH: Water consumption ($/kg)']
                test['LCOH: Finances ($/kg)'] = lcoh_breakdown['LCOH: Finances ($/kg)']



                test = pd.DataFrame(test,index=[0])
                print(test)
                results = pd.concat([results,test])

                #Run pipeline export system financials
                export_hvdc = False
                h2a_solution,h2a_summary,lcoh_breakdown,electrolyzer_installed_cost_kw = run_profast_for_hydrogen.run_profast_for_hydrogen(site_name,electrolyzer_size_mw,H2_Results,\
                    electrolyzer_capex_kw,time_between_replacement,hydrogen_storage_capacity_kg,hydrogen_storage_cost_USDprkg,\
                    desal_capex,desal_opex,useful_life,atb_year,water_cost,wind_size_mw,solar_size_mw, \
                    hybrid_plant,revised_renewable_cost,wind_om_cost_kw,
                    total_h2export_system_cost,
                    opex_pipeline,
                    export_hvdc, combined_pv_wind_storage_power_production_hopp,
                    grid_connected_hopp,h2_ptc,wind_ptc,wind_itc)
                lcoh = h2a_solution['price']
                print('LCOH: ', lcoh)
                # # Max hydrogen production rate [kg/hr]
                max_hydrogen_production_rate_kg_hr = np.max(H2_Results['hydrogen_hourly_production'])
                max_hydrogen_delivery_rate_kg_hr  = np.mean(H2_Results['hydrogen_hourly_production'])

                electrolyzer_capacity_factor = H2_Results['cap_factor']

                test = dict()

                test['Site Name'] = site_name
                test['ATB Year'] = atb_year
                test['HVDC Export'] = export_hvdc
                test['Pipeline Export'] = True
                test['Plant life'] = useful_life
                test['Policy'] = option
                test['Turbine size (MW)'] = turbine_rating
                test['LCOH: total ($/kg)'] = lcoh_breakdown['LCOH: total ($/kg)']
                test['Wind Plant size (MW)'] = wind_size_mw
                test['Wind Plant Size Adjusted for Turbine Rating(MW)'] = wind_plant_size /1000
                test['Electrolyzer size (MW)'] = electrolyzer_size_mw
                test['Load Profile (kW)'] = kw_continuous
                test['Energy to Electrolyzer (kW)'] = np.sum(energy_to_electrolyzer)
                test['Wind capacity factor (%)'] = hybrid_plant.wind.capacity_factor
                test['Electrolyzer capacity factor (%)'] = electrolyzer_capacity_factor
                test['H2 production (kg)'] = hydrogen_annual_production
                test['LCOH ($/kg)'] = lcoh
                test['LCOH: Compression & storage ($/kg)'] = lcoh_breakdown['LCOH: Compression & storage ($/kg)']
                test['LCOH: Electrolyzer CAPEX ($/kg)']= lcoh_breakdown['LCOH: Electrolyzer CAPEX ($/kg)']
                test['LCOH: Desalination CAPEX ($/kg)'] = lcoh_breakdown['LCOH: Desalination CAPEX ($/kg)']
                test['LCOH: HVDC Export CAPEX ($/kg)'] = lcoh_breakdown['LCOH: HVDC Export CAPEX ($/kg)']
                test['LCOH: Pipeline Export CAPEX ($/kg)'] = lcoh_breakdown['LCOH: Pipeline Export CAPEX ($/kg)']
                test['LCOH: HVDC Export FOM ($/kg)'] = lcoh_breakdown['LCOH: HVDC Export FOM ($/kg)']
                test['LCOH: Pipeline Export FOM ($/kg)'] = lcoh_breakdown['LCOH: Pipeline Export FOM ($/kg)']
                test['LCOH: Electrolyzer FOM ($/kg)'] = lcoh_breakdown['LCOH: Electrolyzer FOM ($/kg)']
                test['LCOH: Electrolyzer VOM ($/kg)']=lcoh_breakdown['LCOH: Electrolyzer VOM ($/kg)']
                test['LCOH: Desalination FOM ($/kg)'] = lcoh_breakdown['LCOH: Desalination FOM ($/kg)']
                test['LCOH: Renewable plant ($/kg)'] = lcoh_breakdown['LCOH: Renewable plant ($/kg)']
                test['LCOH: Renewable FOM ($/kg)']= lcoh_breakdown['LCOH: Renewable FOM ($/kg)']
                test['LCOH: Taxes ($/kg)']=lcoh_breakdown['LCOH: Taxes ($/kg)']
                test['LCOH: Water consumption ($/kg)'] = lcoh_breakdown['LCOH: Water consumption ($/kg)']
                test['LCOH: Finances ($/kg)'] = lcoh_breakdown['LCOH: Finances ($/kg)']




                test = pd.DataFrame(test,index=[0])
                print(test)
                results = pd.concat([results,test])

results.to_csv(os.path.join(results_dir, "H2_Analysis_osw_h2.csv"))
print('Done')


import numpy as np
import time
from math import log

class GeneticAlgorithm():
    """a simple genetic algorithm"""

    def __init__(self):

        # inputs
        self.bits = np.array([]) # array of ints same length as design_variables. 
        self.bounds = np.array([]) # array of tuples same length as design_variables
        self.variable_type = np.array([]) # array of strings same length as design_variables ('int' or 'float')
        self.objective_function = None # takes design_variables as an input and outputs the objective values (needs to account for any constraints already)
        self.max_generation = 100
        self.population_size = 0
        self.crossover_rate = 0.1
        self.mutation_rate = 0.01
        self.tol = 1E-6
        self.convergence_iters = 5
        
        # internal variables, you could output some of this info if you wanted
        self.design_variables = np.array([]) # the desgin variables as they are passed into self.objective function
        self.nbits = 0 # the total number of bits in each chromosome
        self.nvars = 0 # the total number of design variables
        self.parent_population = np.array([]) # 2D array containing all of the parent individuals
        self.offspring_population = np.array([]) # 2D array containing all of the offspring individuals
        self.parent_fitness = np.array([]) # array containing all of the parent fitnesses
        self.offspring_fitness = np.array([]) # array containing all of the offspring fitnesses
        self.discretized_variables = {} # a dict of arrays containing all of the discretized design variable

        # outputs
        self.solution_history = np.array([])
        self.optimized_function_value = 0.0
        self.optimized_design_variables = np.array([])


    def initialize_population(self):

        self.parent_population = np.random.randint(0,high=2,size=(self.population_size,self.nbits))
        self.offspring_population = np.zeros_like(self.parent_population)

    
    def initialize_limited(self):
        """initialize the population with only a limited number of ones. Use this if having a full random initialization
        would violate constraints most of the time"""

        n_ones = 1

        self.parent_population = np.zeros((self.population_size,self.nbits),dtype=int)
        for i in range(self.population_size):
            self.parent_population[i][0:n_ones] = 1
            np.random.shuffle(self.parent_population[i])
            # self.parent_population[i][-11:-1] = np.random.randint(0,high=2,size=10)
        
        self.offspring_population = np.zeros_like(self.parent_population)


    def chromosome_2_variables(self,chromosome):  
        """convert the binary chromosomes to design variable values"""      

        first_bit = 0

        float_ind = 0

        for i in range(self.nvars):
            binary_value = 0
            for j in range(self.bits[i]):
                binary_value += chromosome[first_bit+j]*2**j
            first_bit += self.bits[i]

            if self.variable_type[i] == "float":
                self.design_variables[i] = self.discretized_variables["float_var%s"%float_ind][binary_value]
                float_ind += 1

            elif self.variable_type[i] == "int":
                self.design_variables[i] = self.bounds[i][0] + binary_value

    
    def optimize_ga(self,initialize="random",crossover="random",print_progress=True,save_progress=False,start_individual=np.array([])):
        """run the genetic algorithm"""

        print("start GA")
        # determine the number of design variables and initialize
        self.nvars = len(self.variable_type)
        self.design_variables = np.zeros(self.nvars)
        float_ind = 0
        for i in range(self.nvars):
            if self.variable_type[i] == "float":
                ndiscretizations = 2**self.bits[i]
                self.discretized_variables["float_var%s"%float_ind] = np.linspace(self.bounds[i][0],self.bounds[i][1],ndiscretizations)
                float_ind += 1

        # determine the total number of bits
        for i in range(self.nvars):
            if self.variable_type[i] == "int":
                int_range = self.bounds[i][1] - self.bounds[i][0]
                int_bits = int(np.ceil(log(int_range,2)))
                self.bits[i] = int_bits
            self.nbits += self.bits[i]        

        # initialize the population
        print("initialize population")
        if self.population_size%2 == 1:
            self.population_size += 1

        if initialize == "random":
            self.initialize_population()
        elif initialize == "limit":
            self.initialize_limited()

        if len(start_individual) > 0:
            self.parent_population[0][:] = start_individual[:]

        # initialize the fitness arrays
        print("initialize fitness")
        self.parent_fitness = np.zeros(self.population_size)
        self.offspring_fitness = np.zeros(self.population_size)

        # initialize fitness of the parent population
        for i in range(self.population_size):
            self.chromosome_2_variables(self.parent_population[i])
            self.parent_fitness[i] = self.objective_function(self.design_variables)

        converged = False
        ngens = 1
        generation = 1
        difference = self.tol * 10000.0
        self.solution_history = np.zeros(self.max_generation+1)
        self.solution_history[0] = np.min(self.parent_fitness)

        print("start optimization")
        while converged==False and ngens < self.max_generation:
            # crossover
            if crossover=="random":
                self.crossover()
            elif crossover=="chunk":
                self.chunk_crossover()
            elif crossover=="matrix":
                # I haven't tested this very much
                self.matrix_crossover()
            elif crossover=="both":
                # I haven't tested this very much
                self.crossover()
                self.chunk_crossover()

            # mutation
            self.mutate()

            # determine fitness of offspring
            # print("")
            for i in range(self.population_size):
                # if print_progress == True:
                    # print ("\033[A                             \033[A")
                    # print(i)
                self.chromosome_2_variables(self.offspring_population[i])
                self.offspring_fitness[i] = self.objective_function(self.design_variables)

            # rank the total population from best to worst
            total_fitness = np.append(self.parent_fitness,self.offspring_fitness)
            ranked_fitness = np.argsort(total_fitness)[0:int(self.population_size)]

            # take the best. Might switch to some sort of tournament, need to read more about what is better
            # for now I've decided to only keep the best members of the population. I have a large population in 
            # the problems I've run with this so I assume sufficient diversity in the population is maintained from that
            total_population = np.vstack([self.parent_population,self.offspring_population])
            self.parent_population[:,:] = total_population[ranked_fitness,:]
            self.parent_fitness[:] = total_fitness[ranked_fitness]
            
            # store solution history and wrap up generation
            self.solution_history[generation] = np.min(self.parent_fitness)

            if generation > self.convergence_iters:
                difference = self.solution_history[generation-self.convergence_iters] - self.solution_history[generation]
            else:
                difference = 1000
            if abs(difference) <= self.tol:
                converged = True
            
            # shuffle up the order of the population
            shuffle_order = np.arange(1,self.population_size)
            np.random.shuffle(shuffle_order)
            shuffle_order = np.append([0],shuffle_order)
            self.parent_population = self.parent_population[shuffle_order]
            self.parent_fitness = self.parent_fitness[shuffle_order]
            if print_progress==True:
                print(self.parent_fitness[0])

            generation += 1
            ngens += 1

            # save the intermediate progress of the optimization
            if save_progress:
                if ngens%save_progress == 0:
                    file = open('progress.txt', 'w')
                    file.write('Best solution: %s'%(np.min(self.parent_fitness)) + '\n')
                    self.chromosome_2_variables(self.parent_population[np.argmin(self.parent_fitness)])
                    file.write('Design Variables: %s'%(self.design_variables) + '\n')
                    file.close()

        # Assign final outputs
        self.solution_history = self.solution_history[0:ngens]
        self.optimized_function_value = np.min(self.parent_fitness)
        self.chromosome_2_variables(self.parent_population[np.argmin(self.parent_fitness)])
        self.optimized_design_variables = self.design_variables


    def crossover(self):
        # Random crossover

        # set offspring equal to parents
        self.offspring_population[:,:] = self.parent_population[:,:]

        # mate conscutive pairs of parents (0,1),(2,3), ...
        # The population is shuffled so this does not need to be randomized
        for i in range(int(self.population_size/2)):
            # trade bits in the offspring
            crossover_arr = np.random.rand(self.nbits)
            for j in range(self.nbits):
                if crossover_arr[j] < self.crossover_rate:
                    self.offspring_population[2*i][j], self.offspring_population[2*i+1][j] = self.offspring_population[2*i+1][j], self.offspring_population[2*i][j]

    
    def chunk_crossover(self):
        # Organized crossover (maintain portions of the chromosome in order)
        # set offspring equal to parents
        self.offspring_population[:,:] = self.parent_population[:,:]

        # mate conscutive pairs of parents (0,1),(2,3), ...
        # The population is shuffled so this does not need to be randomized
        for i in range(int(self.population_size/2)):
            # trade bits in the offspring
            crossover_loc = int(np.random.rand(1)*self.nbits)
            begin1 = self.offspring_population[2*i][0:crossover_loc]
            end1 = self.offspring_population[2*i][crossover_loc:self.nbits]
            begin2 = self.offspring_population[2*i+1][0:crossover_loc]
            end2 = self.offspring_population[2*i+1][crossover_loc:self.nbits]
            self.offspring_population[2*i] = np.append(begin1,end2)
            self.offspring_population[2*i+1] = np.append(begin2,end1)

    
    def matrix_crossover(self):
        # Haven't tested this very much
        # organize the matrix
        N = int(np.sqrt(len(self.parent_population[0])))
        M1 = np.zeros((N,N))
        M2 = np.zeros((N,N))
        C1 = np.zeros((N,N))
        C2 = np.zeros((N,N))

        for i in range(int(self.population_size/2)):

            rc = np.random.randint(0,high=2)
            crossover_loc = np.random.randint(0,high=N)

            for j in range(N):
                M1[j,:] = self.parent_population[2*i][j*N:(j+1)*N]
                M2[j,:] = self.parent_population[2*i+1][j*N:(j+1)*N]

            if rc == 0:
                C1[0:crossover_loc,:] = M2[0:crossover_loc,:]
                C1[crossover_loc:N,:] = M1[crossover_loc:N,:]
                C2[0:crossover_loc,:] = M1[0:crossover_loc,:]
                C2[crossover_loc:N,:] = M2[crossover_loc:N,:]

            elif rc == 1:
                C1[:,0:crossover_loc] = M2[:,0:crossover_loc]
                C1[:,crossover_loc:N] = M1[:,crossover_loc:N]
                C2[:,0:crossover_loc] = M1[:,0:crossover_loc]
                C2[:,crossover_loc:N] = M2[:,crossover_loc:N]
            
            for j in range(N):
                self.offspring_population[2*i][j*N:(j+1)*N] = C1[j,:]
                self.offspring_population[2*i+1][j*N:(j+1)*N] = C2[j,:]


    def mutate(self):
        # Randomly mutate bits of each chromosome
        for i in range(int(self.population_size)):
            # mutate bits in the offspring
            mutate_arr = np.random.rand(self.nbits)
            for j in range(self.nbits):
                if mutate_arr[j] < self.mutation_rate:
                    self.offspring_population[i][j] = (self.offspring_population[i][j]+1)%2




class GreedyAlgorithm():

    # A couple of discrete optimization algorithms with greedy principles
    def __init__(self):

        # inputs
        self.bits = np.array([]) # array of ints same length as design_variables. 0 signifies an integer design variable
        self.bounds = np.array([]) # array of tuples same length as design_variables
        self.variable_type = np.array([]) # array of strings same length as design_variables ('int' or 'float')
        self.objective_function = None # takes design_variables as an input and outputs the objective values (needs to account for any constraints already)
        
        # internal variables, you could output some of this info if you wanted
        self.design_variables = np.array([])
        self.nbits = 0
        self.nvars = 0
        self.parent_population = np.array([])
        self.offspring_population = np.array([])
        self.parent_fitness = 0.0
        self.offspring_fitness = 0.0
        self.discretized_variables = {} 

        # outputs
        self.solution_history = np.array([])
        self.optimized_function_value = 0.0
        self.optimized_design_variables = np.array([])


    def chromosome_2_variables(self,chromosome):        
        """convert the binary chromosomes to design variable values"""
        first_bit = 0

        float_ind = 0

        for i in range(self.nvars):
            binary_value = 0
            for j in range(self.bits[i]):
                binary_value += chromosome[first_bit+j]*2**j
            first_bit += self.bits[i]

            if self.variable_type[i] == "float":
                self.design_variables[i] = self.discretized_variables["float_var%s"%float_ind][binary_value]
                float_ind += 1

            elif self.variable_type[i] == "int":
                self.design_variables[i] = self.bounds[i][0] + binary_value


    def optimize_greedy(self,initialize="ones"):
        # A simple greedy algorithm. Evaluate the objective after switching each bit
        # one at a time. Keep the switched bit that results in the best improvement.
        # Stop after there is no improvement.

        # determine the number of design variables and initialize
        self.nvars = len(self.variable_type)
        self.design_variables = np.zeros(self.nvars)
        float_ind = 0
        for i in range(self.nvars):
            if self.variable_type[i] == "float":
                ndiscretizations = 2**self.bits[i]
                self.discretized_variables["float_var%s"%float_ind] = np.linspace(self.bounds[i][0],self.bounds[i][1],ndiscretizations)
                float_ind += 1

        # determine the total number of bits
        for i in range(self.nvars):
            if self.variable_type[i] == "int":
                int_range = self.bounds[i][1] - self.bounds[i][0]
                int_bits = int(np.ceil(log(int_range,2)+1))
                self.bits[i] = int_bits
            self.nbits += self.bits[i]
        

        # initialize the fitness
        self.parent_fitness = 0.0
        self.offspring_fitness = 0.0

        # initialize the population
        if initialize == "ones":
            done = False
            while done == False:
                self.parent_population = np.zeros(self.nbits,dtype=int)
                self.parent_population[0] = 1
                np.random.shuffle(self.parent_population)
                # self.parent_population[-11:-1] = np.random.randint(0,high=2,size=10)
                # self.parent_population[-6] = 1
                # self.parent_population[-1] = 1
                self.chromosome_2_variables(self.parent_population)
                self.parent_fitness = self.objective_function(self.design_variables)
                print(self.parent_fitness)
                if self.parent_fitness < 1000.0:
                    done = True
        elif initialize=="zeros":
            self.parent_population = np.zeros(self.nbits,dtype=int)
            # self.parent_population[-6] = 1
            # self.parent_population[-1] = 1
        elif initialize=="random":
            done = False
            while done == False:
                self.parent_population = np.random.randint(0,high=2,size=self.nbits)
                self.chromosome_2_variables(self.parent_population)
                self.parent_fitness = self.objective_function(self.design_variables)
                if self.parent_fitness < 1000.0:
                    done = True

        # initialize the offspring population
        self.offspring_population = np.zeros_like(self.parent_population)
        self.offspring_population[:] = self.parent_population[:]

        # initialize the parent fitness
        self.chromosome_2_variables(self.parent_population)
        self.parent_fitness = self.objective_function(self.design_variables)
        
        # initialize optimization
        self.solution_history = np.array([self.parent_fitness])
        converged = False
        best_population = np.zeros(self.nbits)

        while converged==False:
            # loop through every bit
            best_fitness = self.parent_fitness
            best_population[:] = self.parent_population[:]
            for i in range(self.nbits):
                self.offspring_population[:] = self.parent_population[:]
                self.offspring_population[i] = (self.parent_population[i]+1)%2

                # check the fitness
                self.chromosome_2_variables(self.offspring_population)
                self.offspring_fitness = self.objective_function(self.design_variables)

                # check the performance, see if it is the best so far
                if self.offspring_fitness < best_fitness:
                    best_fitness = self.offspring_fitness
                    best_population[:] = self.offspring_population[:]
            
            # check convergence
            if best_fitness == self.parent_fitness:
                converged = True
            
            # update values if not converged
            else:
                print(best_fitness)
                self.solution_history = np.append(self.solution_history,best_fitness)
                self.parent_population[:] = best_population[:]
                self.parent_fitness = best_fitness

        # final outputs
        self.optimized_function_value = self.parent_fitness
        self.chromosome_2_variables(self.parent_population)
        self.optimized_design_variables = self.design_variables


    def optimize_switch(self,initialize="random",print_progress=True):

        # determine the number of design variables and initialize
        self.nvars = len(self.variable_type)
        self.design_variables = np.zeros(self.nvars)
        float_ind = 0
        for i in range(self.nvars):
            if self.variable_type[i] == "float":
                ndiscretizations = 2**self.bits[i]
                self.discretized_variables["float_var%s"%float_ind] = np.linspace(self.bounds[i][0],self.bounds[i][1],ndiscretizations)
                float_ind += 1

        # determine the total number of bits
        for i in range(self.nvars):
            if self.variable_type[i] == "int":
                int_range = self.bounds[i][1] - self.bounds[i][0]
                int_bits = int(np.ceil(log(int_range,2)+1))
                self.bits[i] = int_bits
            self.nbits += self.bits[i]
        
        init = True
        while init == True:
        # initialize the population
            if initialize == "limit":
                nones = 1
                self.parent_population = np.zeros(self.nbits,dtype=int)
                self.parent_population[0:nones] = 1
                np.random.shuffle(self.parent_population)
            else:
                self.parent_population = np.random.randint(0,high=2,size=self.nbits)
            self.offspring_population = np.zeros_like(self.parent_population)
            self.offspring_population[:] = self.parent_population[:]

            # initialize the fitness
            self.parent_fitness = 0.0
            self.offspring_fitness = 0.0

            self.chromosome_2_variables(self.parent_population)
            self.parent_fitness = self.objective_function(self.design_variables)
            if self.parent_fitness != 1E6:
                init = False
        
        # initialize the optimization
        converged = False
        converged_counter = 0
        self.solution_history = np.array([self.parent_fitness])
        index = 1
        random_method = 0

        # initialize the order array (determines the order of sweeping through the variables)
        order = np.arange(self.nbits)

        last_solution = self.parent_fitness
       
        while converged==False:
            # check if we've gone through every bit
            ind = index%self.nbits
            if ind == 0:
                # check if there has been any change since the last phase
                if last_solution == self.parent_fitness:
                    converged_counter += 1
                else:
                    last_solution = self.parent_fitness
                    converged_counter = 0

                # check convergence
                if converged_counter >= 3:
                    converged = True

                # shuffle the order array and change the phase
                np.random.shuffle(order)
                random_method = (random_method+1)%3
                if print_progress == True:
                    if random_method == 0:
                        print("explore")
                    if random_method == 1:
                        print("switch row")
                    if random_method == 2:
                        print("switch col")

            # set offpring equal to parent

            self.offspring_population[:] = self.parent_population[:]

            # this is the explore phase. Switch a bit, evaluate, and see if we should keep it
            if random_method == 0:
                # switch the value of the appropriate index
                self.offspring_population[order[ind]] = (self.parent_population[order[ind]]+1)%2

                # check the fitness
                self.chromosome_2_variables(self.offspring_population)
                self.offspring_fitness = self.objective_function(self.design_variables)

                # check if we should keep the proposed change
                if self.offspring_fitness < self.parent_fitness:
                    self.solution_history = np.append(self.solution_history,self.offspring_fitness)
                    self.parent_fitness = self.offspring_fitness
                    self.parent_population[:] = self.offspring_population[:]
                    if print_progress == True:
                        print(self.offspring_fitness)

            # this is the first switch phase, switch adjacent bits (only makes sense if they are arranged spatially in a matrix)
            elif random_method == 1:
                # organize the matrix
                N = int(np.sqrt(len(self.parent_population)))
                M = np.zeros((N,N))
                for i in range(N):
                    M[i,:] = self.offspring_population[i*N:(i+1)*N]

                row = order[ind]%N
                col = int(order[ind]/N)

                # switch adjacent numbers
                t1 = M[row][col]
                t2 = M[(row+1)%N][col]

                if t1 != t2:
                    M[row][col] = t2
                    M[(row+1)%N][col] = t1
                    
                    for i in range(N):
                        self.offspring_population[i*N:(i+1)*N] = M[i][:]
                    # check the fitness
                    self.chromosome_2_variables(self.offspring_population)
                    self.offspring_fitness = self.objective_function(self.design_variables)

                    if self.offspring_fitness < self.parent_fitness:
                        self.solution_history = np.append(self.solution_history,self.offspring_fitness)
                        self.parent_fitness = self.offspring_fitness
                        self.parent_population[:] = self.offspring_population[:]
                        if print_progress == True:
                            print(self.offspring_fitness)

            # this is the second switch phase, switch adjacent bits in the other dimension (only makes sense if they are arranged spatially in a matrix)
            elif random_method == 2:
                # organize the matrix
                N = int(np.sqrt(len(self.parent_population)))
                M = np.zeros((N,N))
                for i in range(N):
                    M[i][:] = self.offspring_population[i*N:(i+1)*N]

                row = order[ind]%N
                col = int(order[ind]/N)

                # switch adjacent numbers
                t1 = M[row][col]
                t2 = M[row][(col+1)%N]

                if t1 != t2:
                    M[row][col] = t2
                    M[row][(col+1)%N] = t1
                    
                    for i in range(N):
                        self.offspring_population[i*N:(i+1)*N] = M[i][:]
                    # check the fitness
                    self.chromosome_2_variables(self.offspring_population)
                    self.offspring_fitness = self.objective_function(self.design_variables)

                    if self.offspring_fitness < self.parent_fitness:
                        self.solution_history = np.append(self.solution_history,self.offspring_fitness)
                        self.parent_fitness = self.offspring_fitness
                        self.parent_population[:] = self.offspring_population[:]
                        if print_progress == True:
                            print(self.offspring_fitness)

            # increment the counter
            index += 1

        # final output values
        self.optimized_function_value = self.solution_history[-1]
        self.chromosome_2_variables(self.parent_population)
        self.optimized_design_variables = self.design_variables



if __name__=="__main__":

    def simple_obj(x):
        return x[0]+x[1]

    def rosenbrock_obj(x):
        return (1-x[0])**2 + 100.0*(x[1]-x[0]**2)**2

    def ackley_obj(x):
        p1 = -20.0*np.exp(-0.2*np.sqrt(0.5*(x[0]**2+x[1]**2)))
        p2 = np.exp(0.5*(np.cos(2.*np.pi*x[0]) + np.cos(2.0*np.pi*x[1]))) + np.e + 20.0
        return p1-p2

    def rastrigin_obj(x):
        A = 10.0
        n = len(x)
        tot = 0
        for i in range(n):
            tot += x[i]**2 - A*np.cos(2.0*np.pi*x[i])
        return A*n + tot


    import matplotlib.pyplot as plt

    # from mpl_toolkits.mplot3d import Axes3D
    # X = np.arange(-5, 5, 0.02)
    # Y = np.arange(-5, 5, 0.02)
    # X, Y = np.meshgrid(X, Y)
    # Z = np.zeros_like(X)
    # for i in range(np.shape(Z)[0]):
    #     for j in range(np.shape(Z)[1]):
    #         Z[i][j] = rastrigin_obj(np.array([X[i][j],Y[i][j]]))
    
    # # Plot the surface.
    # fig = plt.figure()
    # ax = fig.gca(projection='3d')
    # surf = ax.plot_surface(X, Y, Z,linewidth=0, antialiased=False)

    # plt.show()

    ga = GeneticAlgorithm()
    ga.bits = np.array([8,8])
    ga.bounds = np.array([(0.0,1.),(0.,1.2)])
    ga.variable_type = np.array(["int","int"])
    ga.population_size = 50
    ga.max_generation = 100
    ga.objective_function = rastrigin_obj
    ga.crossover_rate = 0.1
    ga.mutation_rate = 0.01
    ga.convergence_iters = 25
    ga.tol = 1E-8

    ga.optimize_ga()
    print("optimal function value: ", ga.optimized_function_value)
    print("optimal design variables: ", ga.optimized_design_variables)
    print("nbits: ", ga.nbits)
    plt.plot(ga.solution_history)
    plt.show()
    



import copy
import numpy as np
import matplotlib.pyplot as plt
import os


from hopp.simulation.technologies.sites import SiteInfo
from hopp.simulation.technologies.sites import flatirons_site as sample_site
from hopp.simulation.hopp_interface import HoppInterface
from hopp.simulation.technologies.layout.wind_layout_tools import create_grid


# Function to set up the HOPP model
def setup_hopp(
    hopp_config,
    greenheart_config,
    orbit_config,
    turbine_config,
    floris_config,
    design_scenario,
    wind_cost_results=None,
    show_plots=False,
    save_plots=False,
    output_dir="./output/",
):
    
    if "battery" in hopp_config["technologies"].keys() and \
        ("desired_schedule" not in hopp_config["site"].keys() or hopp_config["site"]["desired_schedule"] == []):
        hopp_config["site"]["desired_schedule"] = [greenheart_config["electrolyzer"]["rating"]]*8760
    hopp_site = SiteInfo(**hopp_config["site"])

    # adjust mean wind speed if desired
    wind_data = hopp_site.wind_resource._data["data"]
    wind_speed = [W[2] for W in wind_data]
    if greenheart_config["site"]["mean_windspeed"]:
        if np.average(wind_speed) != greenheart_config["site"]["mean_windspeed"]:
            wind_speed += greenheart_config["site"]["mean_windspeed"] - np.average(
                wind_speed
            )
            for i in np.arange(0, len(wind_speed)):
                # make sure we don't have negative wind speeds after correction
                hopp_site.wind_resource._data["data"][i][2] = np.maximum(
                    wind_speed[i], 0
                )
    else:
        greenheart_config["site"]["mean_windspeed"] = np.average(wind_speed)

    ################ set up HOPP technology inputs
    
    if hopp_config["site"]["wind"]:
        if hopp_config["technologies"]["wind"]["model_name"] == "floris":
            if design_scenario["wind_location"] == "offshore":
                floris_config["farm"]["layout_x"] = (
                    wind_cost_results.orbit_project.phases[
                        "ArraySystemDesign"
                    ].turbines_x.flatten()
                    * 1e3
                )  # ORBIT gives coordinates in km
                # ORBIT produces nans and must be removed for FLORIS layout
                floris_config["farm"]["layout_x"] = floris_config["farm"]["layout_x"][
                    ~np.isnan(floris_config["farm"]["layout_x"])
                ]
                floris_config["farm"]["layout_y"] = (
                    wind_cost_results.orbit_project.phases[
                        "ArraySystemDesign"
                    ].turbines_y.flatten()
                    * 1e3
                )  # ORBIT gives coordinates in km
                # ORBIT produces nans and must be removed for FLORIS layout
                floris_config["farm"]["layout_y"] = floris_config["farm"]["layout_y"][
                    ~np.isnan(floris_config["farm"]["layout_y"])
                ]
                # remove things from turbine_config file that can't be used in FLORIS and set the turbine info in the floris config file

            if design_scenario["wind_location"] == "onshore":
                grid_position = create_grid(
                    site_shape=hopp_site.polygon,
                    center=hopp_site.polygon.centroid,
                    grid_angle=greenheart_config["site"]["wind_layout"]["grid_angle"],
                    intrarow_spacing=(
                        greenheart_config["site"]["wind_layout"]["row_spacing"]
                        * turbine_config["rotor_diameter"]
                    ),
                    interrow_spacing=(
                        greenheart_config["site"]["wind_layout"]["turbine_spacing"]
                        * turbine_config["rotor_diameter"]
                    ),
                    row_phase_offset=greenheart_config["site"]["wind_layout"][
                        "row_phase_offset"
                    ],
                    max_sites=hopp_config["technologies"]["wind"]["num_turbines"],
                )
                # Extracting xy coordinates
                xy_coordinates = [(point.x, point.y) for point in grid_position]
                floris_config["farm"]["layout_x"] = [point.x for point in grid_position]
                floris_config["farm"]["layout_y"] = [point.y for point in grid_position]

                if (
                    len(floris_config["farm"]["layout_x"])
                    < hopp_config["technologies"]["wind"]["num_turbines"]
                ):
                    raise Exception(
                        "size of site is too small, not all turbines were placed."
                    )

            floris_config["farm"]["turbine_type"] = [
                {
                    x: turbine_config[x]
                    for x in turbine_config
                    if x
                    not in {
                        "turbine_rating",
                        "rated_windspeed",
                        "tower",
                        "nacelle",
                        "blade",
                    }
                }
            ]

            hopp_config["technologies"]["wind"]["turbine_rating_kw"] = turbine_config["turbine_rating"] * 1000
            hopp_config["technologies"]["wind"]["floris_config"] = floris_config

        elif hopp_config["technologies"]["wind"]["model_name"] == "pysam":
            hopp_config["technologies"]["wind"]["turbine_rating_kw"] = turbine_config["turbine_rating"] * 1000.  # convert from MW to kW
            hopp_config["technologies"]["wind"]["hub_height"] = turbine_config["hub_height"]
            hopp_config["technologies"]["wind"]["rotor_diameter"] = turbine_config["rotor_diameter"]

        else:
            raise (
                ValueError(
                    "Wind model '%s' not available. Please choose one of ['floris', 'pysam']" % \
                        (hopp_config["technologies"]["wind"]["model_name"]) )
            )

    # setup hopp interface
    hopp_config_internal = copy.deepcopy(hopp_config)

    if "wave" in hopp_config_internal["technologies"].keys():
        wave_cost_dict = hopp_config_internal["technologies"]["wave"].pop("cost_inputs")

    if "battery" in hopp_config_internal["technologies"].keys():
        hopp_config_internal["site"].update({"desired_schedule": hopp_site.desired_schedule})
        
    hi = HoppInterface(hopp_config_internal)
    hi.system.site = hopp_site

    if "wave" in hi.system.technologies.keys():
        hi.system.wave.create_mhk_cost_calculator(wave_cost_dict)
        
    if show_plots or save_plots:
        # plot wind resource if desired
        print("\nPlotting Wind Resource")
        wind_speed = [W[2] for W in hopp_site.wind_resource._data["data"]]
        plt.figure(figsize=(9, 6))
        plt.plot(wind_speed)
        plt.title(
            "Wind Speed (m/s) for selected location \n Lat:{}, Lon: {} \n Average Wind Speed (m/s) {}".format(
                hopp_config['site']['data']['lat'], hopp_config['site']['data']['lon'], np.round(np.average(wind_speed), decimals=3)
            )
        )

        if show_plots:
            plt.show()
        if save_plots:
            savedir = output_dir + "figures/"
            if not os.path.exists(savedir):
                os.makedirs(savedir)
            plt.savefig(savedir + "average_wind_speed.png", bbox_inches="tight")
        print("\n")

    ################ return all the inputs for hopp
    return hi


# Function to run hopp from provided inputs from setup_hopp()
def run_hopp(hi, project_lifetime, verbose=False):

    hi.simulate(project_life=project_lifetime)

    # store results for later use
    hopp_results = {
        "hopp_interface": hi,
        "hybrid_plant": hi.system,
        "combined_hybrid_power_production_hopp": \
            hi.system.grid._system_model.Outputs.system_pre_interconnect_kwac[0:8760],
        "combined_hybrid_curtailment_hopp": hi.system.grid.generation_curtailed,
        "energy_shortfall_hopp": hi.system.grid.missed_load,
        "annual_energies": hi.system.annual_energies,
        "hybrid_npv": hi.system.net_present_values.hybrid,
        "npvs": hi.system.net_present_values,
        "lcoe": hi.system.lcoe_real,
        "lcoe_nom": hi.system.lcoe_nom,
    }
    if verbose:
        print("\nHOPP Results")
        print("Hybrid Annual Energy: ", hopp_results["annual_energies"])
        print("Capacity factors: ", hi.system.capacity_factors)
        print("Real LCOE from HOPP: ", hi.system.lcoe_real)

    return hopp_results



import pytest
import os
from pathlib import Path

import ORBIT as orbit
from greenheart.simulation.technologies.offshore.fixed_platform import install_platform, calc_platform_opex, calc_substructure_mass_and_cost

'''Sources:
    - [1]  M. Maness, B. Maples and A. Smith, "NREL Offshore Balance-of-System Model," National Renewable Energy Laboratory, 2017. https://www.nrel.gov/docs/fy17osti/66874.pdf
'''
@pytest.mark.skip(reason="no way of currently testing this")
@pytest.fixture
def config():
    offshore_path = Path(__file__).parents[3] / "greenheart" / "simulation" / "technologies" / "offshore"

    return orbit.load_config(os.path.join(offshore_path, "example_fixed_project.yaml"))
@pytest.mark.skip(reason="no way of currently testing this")
def test_install_platform(config):
    '''
    Test the code that calculates the platform installation cost
    [1]: equations (91),(113),(98)
    '''
    distance = 24
    mass = 2100
    area = 500
    
    cost = install_platform(mass, area, distance, install_duration=14)

    assert pytest.approx(cost) == 7200014

def test_calc_substructure_cost(config):
    '''
    Test the code that calculates the CapEx from fixed_platform.py
    [1]: equations (81),(83),(84)
    '''
    topmass = 200
    toparea = 1000
    depth = 45
    
    cost, _ = calc_substructure_mass_and_cost(topmass, toparea, depth)

    assert pytest.approx(cost) == 7640000

def test_calc_substructure_mass(config):
    '''
    Test the code that calculates the CapEx from fixed_platform.py
    [1]: equations (81),(83),(84)
    '''
    topmass = 200
    toparea = 1000
    depth = 45
    
    _,mass = calc_substructure_mass_and_cost(topmass, toparea, depth)

    assert pytest.approx(mass,.1) == 372.02

def test_calc_platform_opex():
    '''
    Test the code that calculates the OpEx from fixed_platform.py
    '''
    capex = 28e6
    opex_rate = 0.01
    cost = calc_platform_opex(capex, opex_rate)
    
    assert pytest.approx(cost) == 28e4




from greenheart.simulation.technologies.hydrogen.h2_storage.on_turbine.on_turbine_hydrogen_storage import PressurizedTower
from pytest import approx
import numpy as np

class TestPressurizedTower():

    def test_frustum(self):
        """
        test static methods for geometry of a frustum
        """

        # try a cone
        D_ref= 15.
        h_ref= 2.1
        V_cone_ref= np.pi/3.*(D_ref/2)**2*h_ref

        V_cone= PressurizedTower.compute_frustum_volume(h_ref, D_ref, 0.0)

        assert V_cone == approx(V_cone_ref)

        # try a cylinder
        D_ref= 45.
        h_ref= 612.3
        V_cyl_ref= np.pi/4.*D_ref**2*h_ref

        V_cyl= PressurizedTower.compute_frustum_volume(h_ref, D_ref, D_ref)

        assert V_cyl == approx(V_cyl_ref)

        # try a frustum by delta of two cones
        D0_ref= 5.7
        h0_ref= 0.0 + 1.0
        D1_ref= 1.9
        h1_ref= 2.4 + 1.0
        D2_ref= 0.0
        h2_ref= 2.4 + 1.2 + 1.0
        V_2_ref= np.pi/3.*(D1_ref/2)**2*(h2_ref - h1_ref)
        V_12_ref= np.pi/3.*(D0_ref/2)**2*(h2_ref - h0_ref)
        V_1_ref= V_12_ref - V_2_ref

        V_1= PressurizedTower.compute_frustum_volume(h1_ref - h0_ref, D0_ref, D1_ref)
        V_12= PressurizedTower.compute_frustum_volume(h2_ref - h0_ref, D0_ref, D2_ref)
        V_2= PressurizedTower.compute_frustum_volume(h2_ref - h1_ref, D1_ref, D2_ref)

        assert V_1 == approx(V_1_ref)
        assert V_12 == approx(V_12_ref)
        assert V_2 == approx(V_2_ref)

    def test_crossover_pressure(self):

        # random plausible values
        E= 0.8
        Sut= 1200e3
        d_t_ratio= 321.

        p_ref= 4*E*Sut/(7*d_t_ratio*(1 - E/7.))

        p= PressurizedTower.get_crossover_pressure(E, Sut, d_t_ratio)

        assert p == approx(p_ref)

    if False: # paper values are untrustworthy
        def test_thickness_increment_const(self):

            # plot values
            p= 600.0e3
            Sut= 636.0e6
            
            alpha_dtp= PressurizedTower.get_thickness_increment_const(p, Sut)

            # values from graph seem to be off by a factor of ten... jacked up!!!
            # these are the values as I read them off the graph assuming that factor
            # is in fact erroneous
            assert 2*1.41*alpha_dtp == approx(0.675e-3, rel= 0.05)
            assert 2*2.12*alpha_dtp == approx(0.900e-3, abs= 0.05)
            assert 2*2.82*alpha_dtp == approx(1.250e-3, abs= 0.05)

    def test_cylinder(self):
        """
        a hypothetical (nonsensical) cylindical tower -> easy to compute
        """

        ### SETUP REFERENCE VALUES

        # input reference values
        h_ref= 100.
        D_ref= 10.
        d_t_ratio_ref= 320.
        density_steel_ref= 7817. # kg/m^3
        strength_ultimate_steel_ref= 636e6 # Pa
        strength_yield_steel_ref= 350e6 # Pa
        Eweld_ref= 0.80
        costrate_steel_ref= 1.50
        costrate_cap_ref= 2.66
        costrate_ladder_ref= 32.80
        cost_door_ref= 2000.
        cost_mainframe_ref= 6300
        cost_nozzlesmanway_ref= 16000
        costrate_conduit_ref= 35
        temp_ref= 25. # degC
        R_H2_ref= 4126. # J/(kg K)
        maintenance_rate_ref= 0.03
        staff_hours_ref= 60
        wage_ref= 36

        # geometric reference values
        thickness_wall_trad_ref= D_ref/d_t_ratio_ref
        surfacearea_wall_ref= np.pi*D_ref*h_ref
        surfacearea_cap_ref= np.pi/4.*D_ref**2

        # non-pressurized/traditional geometry values
        volume_wall_trad_ref= surfacearea_wall_ref*thickness_wall_trad_ref
        volume_inner_ref= h_ref*surfacearea_cap_ref
        volume_cap_top_trad_ref= 0.0 # surfacearea_cap_ref*thickness_top_ref
        volume_cap_bot_trad_ref= 0.0 # surfacearea_cap_ref*thickness_bot_ref

        # non-pressurized/traditional mass/cost values
        mass_wall_trad_ref= density_steel_ref*volume_wall_trad_ref
        mass_cap_top_trad_ref= density_steel_ref*volume_cap_top_trad_ref
        mass_cap_bot_trad_ref= density_steel_ref*volume_cap_bot_trad_ref
        cost_tower_trad_ref= costrate_steel_ref*(mass_wall_trad_ref + mass_cap_top_trad_ref + mass_cap_bot_trad_ref)
        cost_nontower_trad_ref= h_ref*costrate_ladder_ref + cost_door_ref

        # pressurization info
        p_crossover_ref= 4*Eweld_ref*strength_ultimate_steel_ref/(7*d_t_ratio_ref*(1 - Eweld_ref/7))
        delta_t_ref= p_crossover_ref*(D_ref/2)/(2*strength_ultimate_steel_ref)
        thickness_wall_ref= D_ref/d_t_ratio_ref + delta_t_ref
        thickness_cap_top_ref= D_ref*np.sqrt(0.10*p_crossover_ref/(Eweld_ref*strength_yield_steel_ref/1.5))
        thickness_cap_bot_ref= D_ref*np.sqrt(0.10*p_crossover_ref/(Eweld_ref*strength_yield_steel_ref/1.5))

        # pressurized geometry values
        volume_wall_ref= surfacearea_wall_ref*thickness_wall_ref
        volume_cap_top_ref= surfacearea_cap_ref*(thickness_cap_top_ref)
        volume_cap_bot_ref= surfacearea_cap_ref*(thickness_cap_bot_ref)

        # pressurized mass/cost values
        mass_wall_ref= density_steel_ref*volume_wall_ref
        mass_cap_top_ref= density_steel_ref*volume_cap_top_ref
        mass_cap_bot_ref= density_steel_ref*volume_cap_bot_ref
        cost_tower_ref= costrate_steel_ref*mass_wall_ref + costrate_cap_ref*(mass_cap_top_ref + mass_cap_bot_ref)
        cost_nontower_ref= 2*h_ref*costrate_ladder_ref + 2*cost_door_ref + cost_mainframe_ref + cost_nozzlesmanway_ref + costrate_conduit_ref*h_ref

        # gas
        rho_H2_ref= p_crossover_ref/(R_H2_ref*(temp_ref + 273.15))
        m_H2_ref= volume_inner_ref*rho_H2_ref

        # capex
        capex_ref= cost_tower_ref + cost_nontower_ref - cost_tower_trad_ref - cost_nontower_trad_ref

        # opex
        opex_ref= maintenance_rate_ref*capex_ref + wage_ref*staff_hours_ref

        turbine= {
            'tower_length': h_ref,
            'section_diameters': [D_ref, D_ref, D_ref],
            'section_heights': [0., 0. + 0.5*h_ref, 0. + h_ref],
            # 'section_diameters': [D_ref, D_ref],
            # 'section_heights': [0., 0. + h_ref],
        }

        ## traditional estimates (non-pressurized)

        pressurized_cylinder= PressurizedTower(1992, turbine)

        assert pressurized_cylinder.get_volume_tower_inner() == approx(volume_inner_ref)
        assert pressurized_cylinder.get_volume_tower_material(pressure= 0)[0] == approx(volume_wall_trad_ref)
        assert pressurized_cylinder.get_volume_tower_material(pressure= 0)[1] == approx(volume_cap_bot_trad_ref)
        assert pressurized_cylinder.get_volume_tower_material(pressure= 0)[2] == approx(volume_cap_top_trad_ref)
        assert pressurized_cylinder.get_mass_tower_material(pressure= 0)[0] == approx(mass_wall_trad_ref)
        assert pressurized_cylinder.get_mass_tower_material(pressure= 0)[1] == approx(mass_cap_bot_trad_ref)
        assert pressurized_cylinder.get_mass_tower_material(pressure= 0)[2] == approx(mass_cap_top_trad_ref)
        
        assert np.sum(pressurized_cylinder.get_cost_tower_material(pressure= 0)) == approx(cost_tower_trad_ref)
        assert pressurized_cylinder.get_cost_nontower(traditional= True) == approx(cost_nontower_trad_ref)
        
        ## pressurized estimates

        assert pressurized_cylinder.operating_pressure == p_crossover_ref

        assert pressurized_cylinder.get_volume_tower_material()[0] == approx(volume_wall_ref)
        assert pressurized_cylinder.get_volume_tower_material()[1] == approx(volume_cap_bot_ref)
        assert pressurized_cylinder.get_volume_tower_material()[2] == approx(volume_cap_top_ref)
        assert pressurized_cylinder.get_mass_tower_material()[0] == approx(mass_wall_ref)
        assert pressurized_cylinder.get_mass_tower_material()[1] == approx(mass_cap_bot_ref)
        assert pressurized_cylinder.get_mass_tower_material()[2] == approx(mass_cap_top_ref)
        
        assert np.sum(pressurized_cylinder.get_cost_tower_material()) == approx(cost_tower_ref)
        assert pressurized_cylinder.get_cost_nontower() == approx(cost_nontower_ref)

        ## output interface

        # make sure the final values match expectation
        assert pressurized_cylinder.get_capex() == approx(capex_ref)
        assert pressurized_cylinder.get_opex() == approx(opex_ref)
        assert pressurized_cylinder.get_mass_empty() == approx(mass_wall_ref + mass_cap_bot_ref + mass_cap_top_ref
                                                               - mass_wall_trad_ref - mass_cap_bot_trad_ref - mass_cap_top_trad_ref)
        assert pressurized_cylinder.get_capacity_H2() == approx(m_H2_ref)
        assert pressurized_cylinder.get_pressure_H2() == approx(p_crossover_ref)


    if True:
        def test_cone(self):
            """
            a hypothetical (nonsensical) conical tower -> easy to compute
            """

            ### SETUP REFERENCE VALUES

            # input reference values
            h_ref= 81.
            D_base_ref= 10.
            D_top_ref= 0.

            # non-input parameters
            d_t_ratio_ref= 320.
            density_steel_ref= 7817. # kg/m^3
            strength_ultimate_steel_ref= 636e6 # Pa
            strength_yield_steel_ref= 350e6 # Pa
            Eweld_ref= 0.8
            costrate_steel_ref= 1.50
            costrate_cap_ref= 2.66
            costrate_ladder_ref= 32.80
            cost_door_ref= 2000.
            cost_mainframe_ref= 6300
            cost_nozzlesmanway_ref= 16000
            costrate_conduit_ref= 35
            temp_ref= 25. # degC
            R_H2_ref= 4126. # J/(kg K)
            maintenance_rate_ref= 0.03
            staff_hours_ref= 60
            wage_ref= 36

            # geometric reference values
            surfacearea_cap_top_ref= np.pi/4.*D_top_ref**2
            surfacearea_cap_bot_ref= np.pi/4.*D_base_ref**2
            thickness_wall_top_ref= D_top_ref/d_t_ratio_ref
            thickness_wall_bot_ref= D_base_ref/d_t_ratio_ref
            
            def cone_volume(h, d):
                return np.pi/3.*(d/2)**2*h

            # non-pressurized/traditional geometry values
            volume_inner_ref= cone_volume(h_ref, D_base_ref)
            print(volume_inner_ref)
            volume_wall_trad_ref= cone_volume(h_ref, D_base_ref + thickness_wall_bot_ref) \
                    - cone_volume(h_ref, D_base_ref - thickness_wall_bot_ref)
            volume_cap_top_trad_ref= 0.0 # surfacearea_cap_top_ref*thickness_top_ref
            volume_cap_bot_trad_ref= 0.0 # surfacearea_cap_bot_ref*thickness_bot_ref

            # non-pressurized/traditional mass/cost values
            mass_wall_trad_ref= density_steel_ref*volume_wall_trad_ref
            mass_cap_top_trad_ref= density_steel_ref*volume_cap_top_trad_ref
            mass_cap_bot_trad_ref= density_steel_ref*volume_cap_bot_trad_ref
            cost_tower_trad_ref= costrate_steel_ref*(mass_wall_trad_ref + mass_cap_top_trad_ref + mass_cap_bot_trad_ref)
            cost_nontower_trad_ref= h_ref*costrate_ladder_ref + cost_door_ref

            # pressurization info
            p_crossover_ref= 4*Eweld_ref*strength_ultimate_steel_ref/(7*d_t_ratio_ref*(1 - Eweld_ref/7))
            dt_bot_ref= p_crossover_ref*(D_base_ref/2)/(2*strength_ultimate_steel_ref)
            thickness_wall_bot_ref= D_base_ref/d_t_ratio_ref + dt_bot_ref
            thickness_cap_top_ref= D_top_ref*np.sqrt(0.10*p_crossover_ref/(Eweld_ref*strength_yield_steel_ref/1.5))
            thickness_cap_bot_ref= D_base_ref*np.sqrt(0.10*p_crossover_ref/(Eweld_ref*strength_yield_steel_ref/1.5))

            # pressurized geometry values
            volume_wall_ref= cone_volume(h_ref, D_base_ref + thickness_wall_bot_ref) \
                    - cone_volume(h_ref, D_base_ref - thickness_wall_bot_ref)
            volume_cap_top_ref= surfacearea_cap_top_ref*(thickness_cap_top_ref)
            volume_cap_bot_ref= surfacearea_cap_bot_ref*(thickness_cap_bot_ref)

            # pressurized mass/cost values
            mass_wall_ref= density_steel_ref*volume_wall_ref
            mass_cap_top_ref= density_steel_ref*volume_cap_top_ref
            mass_cap_bot_ref= density_steel_ref*volume_cap_bot_ref
            cost_tower_ref= costrate_steel_ref*mass_wall_ref + costrate_cap_ref*(mass_cap_top_ref + mass_cap_bot_ref)
            cost_nontower_ref= 2*h_ref*costrate_ladder_ref + 2*cost_door_ref + cost_mainframe_ref \
                    + cost_nozzlesmanway_ref + costrate_conduit_ref*h_ref

            # gas
            rho_H2_ref= p_crossover_ref/(R_H2_ref*(temp_ref + 273.15))
            m_H2_ref= volume_inner_ref*rho_H2_ref

            # capex
            capex_ref= cost_tower_ref + cost_nontower_ref - cost_tower_trad_ref - cost_nontower_trad_ref

            # opex
            opex_ref= maintenance_rate_ref*capex_ref + wage_ref*staff_hours_ref

            turbine= {
                'tower_length': h_ref,
                # 'section_diameters': [D_base_ref, D_top_ref],
                # 'section_heights': [0., 0. + h_ref],
                'section_diameters': [D_base_ref, 0.5*(D_top_ref + D_base_ref), D_top_ref],
                'section_heights': [0., 0. + h_ref/2., 0. + h_ref],
            }

            ## traditional estimates (non-pressurized)

            pressurized_cone= PressurizedTower(1992, turbine)

            assert pressurized_cone.get_volume_tower_inner() == approx(volume_inner_ref)
            assert pressurized_cone.get_volume_tower_material(pressure= 0)[0] == approx(volume_wall_trad_ref)
            assert pressurized_cone.get_volume_tower_material(pressure= 0)[1] == approx(volume_cap_bot_trad_ref)
            assert pressurized_cone.get_volume_tower_material(pressure= 0)[2] == approx(volume_cap_top_trad_ref)
            assert pressurized_cone.get_mass_tower_material(pressure= 0)[0] == approx(mass_wall_trad_ref)
            assert pressurized_cone.get_mass_tower_material(pressure= 0)[1] == approx(mass_cap_bot_trad_ref)
            assert pressurized_cone.get_mass_tower_material(pressure= 0)[2] == approx(mass_cap_top_trad_ref)
            
            assert np.sum(pressurized_cone.get_cost_tower_material(pressure= 0)) == approx(cost_tower_trad_ref)
            assert pressurized_cone.get_cost_nontower(traditional= True) == approx(cost_nontower_trad_ref)
            
            ## pressurized estimates

            assert pressurized_cone.operating_pressure == p_crossover_ref

            assert pressurized_cone.get_volume_tower_material()[0] == approx(volume_wall_ref)
            assert pressurized_cone.get_volume_tower_material()[1] == approx(volume_cap_bot_ref)
            assert pressurized_cone.get_volume_tower_material()[2] == approx(volume_cap_top_ref)
            assert pressurized_cone.get_mass_tower_material()[0] == approx(mass_wall_ref)
            assert pressurized_cone.get_mass_tower_material()[1] == approx(mass_cap_bot_ref)
            assert pressurized_cone.get_mass_tower_material()[2] == approx(mass_cap_top_ref)
            
            assert np.sum(pressurized_cone.get_cost_tower_material()) == approx(cost_tower_ref)
            assert pressurized_cone.get_cost_nontower() == approx(cost_nontower_ref)

            ## output interface

            # make sure the final values match expectation
            assert pressurized_cone.get_capex() == approx(capex_ref)
            assert pressurized_cone.get_opex() == approx(opex_ref)
            assert pressurized_cone.get_mass_empty() == approx(mass_wall_ref + mass_cap_bot_ref + mass_cap_top_ref
                                                                - mass_wall_trad_ref - mass_cap_bot_trad_ref
                                                                - mass_cap_top_trad_ref)
            assert pressurized_cone.get_capacity_H2() == approx(m_H2_ref)
            assert pressurized_cone.get_pressure_H2() == approx(p_crossover_ref)

    if True:
        def test_paper(self):
        
            h_ref= 84.
            D_bot_ref= 5.66
            D_top_ref= 2.83
            # d_t_ratio_ref= 320.
            # rho_density_ref= 7817
            # costrate_steel= 1.50
            cost_tower_ref= 183828

            cost_tower_trad_ref= 183828
            cost_nontower_trad_ref= 188584 - cost_tower_trad_ref
            m_H2_stored_ref= 951 # kg
            cost_tower_ref= cost_tower_trad_ref + 21182
            cost_cap_bot_ref= 29668
            cost_cap_top_ref= 5464
            cost_nontower_ref= 2756 + 2000 + 2297 + 2450 + 6300 + 15918

            turbine= {
                'tower_length': 84.,
                'section_diameters': [5.66, 4.9525, 4.245, 3.5375, 2.83],
                'section_heights': [0., 21., 42., 63., 84.]
            }
        
            pressurized_tower_instance= PressurizedTower(2004, turbine)
            pressurized_tower_instance.run()
            
            Vinner_balpark = PressurizedTower.compute_frustum_volume(h_ref, D_bot_ref, D_top_ref)
            
            # traditional sizing should get cost within 5%
            assert pressurized_tower_instance.get_cost_tower_material(pressure= 0)[0] == \
                    approx(cost_tower_trad_ref, rel= 0.05)
            assert pressurized_tower_instance.get_cost_tower_material(pressure= 0)[1] == 0.0
            assert pressurized_tower_instance.get_cost_tower_material(pressure= 0)[2] == 0.0
            assert pressurized_tower_instance.get_cost_nontower(traditional= True) == \
                    approx(cost_nontower_trad_ref, rel= 0.05)

            # pressurized sizing should get wall cost within 10%
            assert pressurized_tower_instance.get_cost_tower_material()[0] == approx(cost_tower_ref, rel= 0.10)
            # not sure why but the cap sizing is way off: 200% error allowed for bottom cap
            assert pressurized_tower_instance.get_cost_tower_material()[1] == approx(cost_cap_bot_ref, rel= 2.0)
            # not sure why but the cap sizing is way off: 100% error allowed for top cap
            assert pressurized_tower_instance.get_cost_tower_material()[2] == approx(cost_cap_top_ref, rel= 1.0)

            # non-tower pressurized sizing evidently has some weird assumptions but should get within 10%
            assert pressurized_tower_instance.get_cost_nontower() == approx(cost_nontower_ref, rel= 0.1)

            # capacity within 10%
            assert pressurized_tower_instance.get_capacity_H2() == approx(m_H2_stored_ref, rel= 0.1)

if __name__ == "__main__":
    test_set= test_pressurized_tower()
    #test_set= TestPressurizedTower()



# -*- coding: utf-8 -*-
"""
Created on Fri Jan 15 15:06:21 2021

@author: ppeng
"""
import openpyxl as openpyxl
import numpy as np
import math as math
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from CoolProp.CoolProp import PropsSI
from scipy.optimize import leastsq

plt.rcParams.update({'font.size': 13})

class CompressedGasFunction():

    def __init__(self, path_tankinator):
        # path to the excel spreadsheet to store material properties
        self.wb_tankinator = openpyxl.load_workbook(path_tankinator, data_only=True) #Add file name
        
        ################Other key inputs besides the main script##########################
        self.MW_H2=2.02e-03 #molecular weight of H2 in kg/mol
        
        self.Pres=int(350) #Define storage pressure in bar, Important, if you change storage pressure, make sure to change it in the corresponding tab in Tankinator and save again
        self.Temp_c = 293 #Define storage temperature in K
        self.Pin=int(30)  #Deinfe pressure out of electrolyzer in bar
        self.Tin=int(353) #Define temperature out of electrolyzer in K
        self.T_amb=int(295)
        self.Pres3=int(35) #Define outlet pressure in bar
        self.Temp3=int(353) #Define outlet temperature in K
        
        self.start_point = 10   #For setting the smallest capacity for fitting and plotting

        
        #################Economic parameters
        self.CEPCI2007 = 525.4
        self.CEPCI2001 = 397
        self.CEPCI2017 = 567.5
        
        self.CEPCI_current = 708    ####Change this value for current CEPCI 
        
        self.wage = 36
        self.maintanance = 0.03
        self.Site_preparation = 100   #Site preparation in $/kg
        
        self.Tank_manufacturing = 1.8 #self.Markup for tank manufacturing
        self.Markup = 1.5   #self.Markup for installation engineering/contingency
        
        #################Other minor input parameters############
        self.R=8.314 # gas onstant m3*Pa/(molK)
        self.Heat_Capacity_Wall = 0.92 ##wall heat capacity at 298 K in kJ/kg*K for carbon fiber composite
        self.Efficiency_comp = 0.7  #Compressor efficiency
        self.Efficiency_heater = 0.7  #Heat efficiency

    def exp_log_fit(self, var_op, capacity_1):
        
        a_op=var_op[0]
        b_op=var_op[1]
        c_op=var_op[2]

        fit_op_kg = np.exp(a_op*(np.log(capacity_1))**2-b_op*np.log(capacity_1)+c_op)

        return fit_op_kg

    def residual_op(self, var_op, capacity_1, Op_c_Costs_kg):
        
            fit_op_kg = self.exp_log_fit(var_op, capacity_1)

            return (fit_op_kg - Op_c_Costs_kg)
    
    def exp_fit(self, x, a, b):
            return a*x**b

    def calculate_max_storage_capacity(self, Wind_avai, H2_flow, Release_efficiency):
        
        H2_flow_ref = 200 #reference flow rate of steel plants in tonne/day, in case in the future it is not 200 tonne/day
        
        capacity_max = (0.8044*Wind_avai**2-57.557*Wind_avai+4483.1)*(H2_flow/H2_flow_ref)/Release_efficiency*1000  ###Total max equivalent storage capacity kg

        return capacity_max

    def calculate_max_storage_duration(self, Release_efficiency, H2_flow):

        t_discharge_hr_max = self.capacity_max/1000*Release_efficiency/H2_flow  ###This is the theoretical maximum storage duration 

        return t_discharge_hr_max
    
    #TODO keep breaking this up so we can run the model without running the curve fit
    def func(self, Wind_avai, H2_flow, cdratio, Energy_cost, cycle_number, capacity_max_spec=None, t_discharge_hr_max_spec=None):
        """
        Run the compressor and storage container cost models

        Wind_avai is only used for calculating the theoretical maximum storage capacity prior to curve fitting

        H2_flow is (I think) the rate the H2 is being removed from the tank in Tonne/day

        cdratio is the charge/discharge ratio (1 means charge rate equals the discharge rate, 2 means charge is 2x the discharge rate)

        Energy_cost is the renewable energy cost in $/kWh, or can be set to 0 to exclude energy costs

        cycle number should just be left as 1 (see compressed_all.py)
        """
        
        ##############Calculation of storage capacity from duration#############
        if 1-self.Pres3/self.Pres < 0.9:
            Release_efficiency = 1-self.Pres3/self.Pres    
        else:
            Release_efficiency = 0.9
        
        if capacity_max_spec == None:
            self.capacity_max = self.calculate_max_storage_capacity(Wind_avai, H2_flow, Release_efficiency)
        else:
            self.capacity_max = capacity_max_spec
        
        if t_discharge_hr_max_spec == None:
            self.t_discharge_hr_max = self.calculate_max_storage_duration(Release_efficiency, H2_flow)
        else:
            self.t_discharge_hr_max = t_discharge_hr_max_spec

        if self.verbose:
            print('Maximum capacity is', self.capacity_max, 'kg H2')
            print('Maximum storage duration is', self.t_discharge_hr_max, 'hr')
        
        if self.Pres > 170:
        ####Use this if use type IV tanks
            tank_type= 4
            sheet_tankinator = self.wb_tankinator['type4_rev3'] #Add Sheet name
            Vtank_c_cell = sheet_tankinator.cell(row=19, column=3)   #tank internal volume in cm3
            Vtank_c=Vtank_c_cell.value/(10**6) #tank volume in m3
            m_c_wall_cell=sheet_tankinator.cell(row=55, column=3)
            m_c_wall=m_c_wall_cell.value #Wall mass in kg
            Mtank_c=m_c_wall #TODO why is this set but not used?
            Louter_c_cell=sheet_tankinator.cell(row= 36, column= 3)
            length_outer_c=Louter_c_cell.value # outer length of tank
            Router_c_cell=sheet_tankinator.cell(row= 37, column= 3)
            radius_outer_c=Router_c_cell.value # outer radius of tank
            Cost_c_tank_cell=sheet_tankinator.cell(row=65, column=3) #Cost of one tank 
            Cost_c_tank = Cost_c_tank_cell.value   ##Cost of the tank in $/tank
            
        if self.Pres <= 170:
        ####Use this if use type I tanks
            tank_type= 1
            sheet_tankinator = self.wb_tankinator['type1_rev3'] #Add Sheet nam
            Vtank_c_cell = sheet_tankinator.cell(row=20, column=3)  ##Tank's outer volume in cm^3
            Vtank_c=Vtank_c_cell.value/(10**6) #tank volume in m3
            m_c_wall_cell=sheet_tankinator.cell(row=188, column=3)
            m_c_wall=m_c_wall_cell.value #Wall mass in kg
            Mtank_c=m_c_wall #TODO why is this set but not used?
            Louter_c_cell=sheet_tankinator.cell(row= 184, column= 3)
            length_outer_c=Louter_c_cell.value
            Router_c_cell=sheet_tankinator.cell(row= 185, column= 3)
            radius_outer_c=Router_c_cell.value
            Cost_c_tank_cell=sheet_tankinator.cell(row=193, column=3) #Cost of one tank 
            Cost_c_tank = Cost_c_tank_cell.value   ##Cost of the tank in $/tank

        self.tank_type= tank_type
        self.Vtank= Vtank_c
        self.m_H2_tank= self.Vtank*PropsSI("D", "P", self.Pres*10**5, "T", self.Temp_c, "Hydrogen")
        self.Mempty_tank= Mtank_c
        self.Router= radius_outer_c
        self.Louter= length_outer_c

        #####Define arrays for plotting and fitting  

        self.t_discharge_hr_1 = np.linspace (self.t_discharge_hr_max, self.t_discharge_hr_max/self.start_point, num=15)
        self.cost_kg = np.zeros(len(self.t_discharge_hr_1))
        cost_kg_tank = np.zeros(len(self.t_discharge_hr_1))
        cost_kg_comp = np.zeros(len(self.t_discharge_hr_1))
        cost_kg_ref = np.zeros(len(self.t_discharge_hr_1))
        cost_kg_heat = np.zeros(len(self.t_discharge_hr_1))
        self.number_of_tanks = np.zeros(len(self.t_discharge_hr_1))
        self.capacity_1 = np.zeros(len(self.t_discharge_hr_1))
        self.Op_c_Costs_kg = np.zeros(len(self.t_discharge_hr_1))
        self.total_energy_used_kwh = np.zeros(len(self.t_discharge_hr_1))
        
        ###################################################################################################
        ###################################################################################################
        ###################################################################################################
        ###############Starting detailed calculations#################################
        ###############Stage 1 calculations#################################
        
        for i in range (0,len(self.t_discharge_hr_1-1)):
            t_discharge_hr = self.t_discharge_hr_1 [i]
            capacity=H2_flow*t_discharge_hr*1000/Release_efficiency #Maximum capacity in kg H2
            
            self.capacity_1 [i] = capacity
            
            rgas=PropsSI("D", "P", self.Pres*10**5, "T", self.Temp_c, "Hydrogen") #h2 density in kg/m3 under storage conditions
            H2_c_mass_gas_tank = Vtank_c*rgas  #hydrogen mass per tank in kg
            H2_c_mass_tank = H2_c_mass_gas_tank  #Estimation of H2 amount per tank in kg
            self.single_tank_h2_capacity_kg = H2_c_mass_tank
            
            number_c_of_tanks = np.ceil(capacity/H2_c_mass_tank)
            self.number_of_tanks[i]= number_c_of_tanks
            H2_c_Cap_Storage= H2_c_mass_tank*(number_c_of_tanks-1)+capacity%H2_c_mass_tank  ####This will be useful when changing to assume all tanks are full, but will cause the model to not perform well for small scales, where 1 tank makes a large difference
            
            #################Energy balance for adsorption (state 1 to state 2)########
            self.t_charge_hr=t_discharge_hr * (1/cdratio)
            t_precondition_hr=self.t_charge_hr  #correcting first cycle, useful to size based on maximum power and also when calculating the operational cost
            m_c_flow_rate_1_2 = H2_c_Cap_Storage/t_precondition_hr/3600 #mass flow rate in kg/s
            Temp2=self.Temp_c
            Temp1_gas=self.Tin
            Temp1_solid=self.T_amb
            Pres2=self.Pres*10**5
            Pres1=self.Pin*10**5
            H_c_1_spec_g=PropsSI("H", "P", Pres1, "T", Temp1_gas, "Hydrogen")/1000 #specific enthalpy of the gas under T1 P1 in kJ/kg
            H_c_2_spec_g=PropsSI("H", "P", Pres2, "T", Temp2, "Hydrogen")/1000 #specific enthalpy of the gas under T2 P2 in kJ/kg
            H_c_1_gas = H2_c_Cap_Storage*H_c_1_spec_g
            H_c_2_gas = H2_c_Cap_Storage*H_c_2_spec_g
            deltaE_c_H2_1_2 = H_c_2_gas-H_c_1_gas
            deltaE_c_Uwall_1_2 = self.Heat_Capacity_Wall*(Temp2-Temp1_solid)*m_c_wall*number_c_of_tanks #Net energy/enthalpy change of adsorbent in kJ
            deltaE_c_net_1_2 = deltaE_c_H2_1_2 + deltaE_c_Uwall_1_2 #Net energy/enthalpy change in kJ
            deltaP_c_net_1_2 = deltaE_c_net_1_2/self.t_charge_hr/3600  #Net power change in kW
                
            #################Energy balance for desorption (state 2 to state 3)########
            Temp3_gas=self.Temp3
            Temp3_solid = Temp2
            self.Pres3=self.Pres3
            Pres3_tank=self.Pres*(1-Release_efficiency)*10**5*10
            H_c_3_spec_g_fuel_cell=PropsSI("H", "P", self.Pres3, "T", Temp3_gas, "Hydrogen")/1000 #specific enthalpy of the released gas in kJ/kg
            H_c_3_spec_g_tank=PropsSI("H", "P", Pres3_tank, "T", Temp2, "Hydrogen")/1000 #specific enthalpy of the remaining free volume gas in kJ/kg
            H_c_3_gas = H2_c_Cap_Storage*Release_efficiency*H_c_3_spec_g_fuel_cell+H2_c_Cap_Storage*(1-Release_efficiency)*H_c_3_spec_g_tank  #Total gas phase enthalpy in stage 3 in kJ
            deltaE_c_H2_2_3=H_c_3_gas-H_c_2_gas #Total h2 enthalpy change in kJ
            deltaE_c_Uwall_2_3 = self.Heat_Capacity_Wall*(Temp3_solid-Temp2)*m_c_wall*number_c_of_tanks  #kJ
            deltaE_c_net_2_3 = deltaE_c_H2_2_3+deltaE_c_Uwall_2_3 # Net enthalpy change during desorption
            detlaP_c_net_2_3 = deltaE_c_net_2_3/t_discharge_hr/3600
            
            ###############Energy balance for adsorption (state 4 to state 2)##########
            m_c_flow_rate_4_2 = H2_c_Cap_Storage*Release_efficiency/self.t_charge_hr/3600
            Temp4_tank=Temp2
            Temp4_gas = self.Tin #TODO why set but not used?
            Pres4=self.Pres3    #TODO why set but not used?
            Pres4_tank = Pres3_tank
            H_c_4_spec_g_electrolyzer=PropsSI("H", "P", self.Pin, "T", self.Tin, "Hydrogen")/1000 #specific enthalpy of the released gas in kJ/kg
            H_c_4_spec_g_tank=PropsSI("H", "P", Pres4_tank, "T", Temp2-5, "Hydrogen")/1000 #specific enthalpy of the remaining free volume gas in kJ/kg
            H_c_4_gas = H2_c_Cap_Storage*Release_efficiency*H_c_4_spec_g_electrolyzer+H2_c_Cap_Storage*(1-Release_efficiency)*H_c_4_spec_g_tank  #Total gas phase enthalpy in stage 3 in kJ
            deltaE_c_H2_4_2=H_c_2_gas-H_c_4_gas #Total h2 enthalpy change in kJ
            deltaE_c_Uwall_4_2 = self.Heat_Capacity_Wall*(Temp2-Temp4_tank)*m_c_wall*number_c_of_tanks  #kJ
            deltaE_c_net_4_2 = deltaE_c_H2_4_2 +deltaE_c_Uwall_4_2 # Net enthalpy change during desorption
            deltaP_c_net_4_2 = deltaE_c_net_4_2/self.t_charge_hr/3600
            
            
            ########################################Costs for cycle 1 adsorption##################################
            
            ########################################CAPITAL COSTS (sized based on cycle 1 requirements)###########################################
            
            ###############################Compressor costs ### axial/centrifugal
            if self.Pres>=self.Pin:
                K=PropsSI("ISENTROPIC_EXPANSION_COEFFICIENT", "P", self.Pin*10**5, "T", self.Tin, "Hydrogen")
                P2nd = self.Pin*(self.Pres/self.Pin)**(1/3)
                P3rd = self.Pin*(self.Pres/self.Pin)**(1/3)*(self.Pres/self.Pin)**(1/3)
                work_c_comp_1 = K/(K-1)*self.R*self.Tin/self.MW_H2*((P2nd/self.Pin)**((K-1)/K)-1)
                work_c_comp_2 = K/(K-1)*self.R*self.Tin/self.MW_H2*((P3rd/P2nd)**((K-1)/K)-1)
                work_c_comp_3 = K/(K-1)*self.R*self.Tin/self.MW_H2*((self.Pres/P3rd)**((K-1)/K)-1)
                Work_c_comp = work_c_comp_1+work_c_comp_2+work_c_comp_3
                # Work_c_comp=K/(K-1)*self.R*self.Tin/self.MW_H2*((self.Pres/self.Pin)**((K-1)/K)-1) #mechanical energy required for compressor in J/kg (single stage)
                Power_c_comp_1_2=Work_c_comp/1000*m_c_flow_rate_1_2 #mechanical power of the pump in kW
                Power_c_comp_4_2=Work_c_comp/1000*m_c_flow_rate_4_2
                A_c_comp_1_2 = Power_c_comp_1_2/self.Efficiency_comp  #total power in kW
                A_c_comp_4_2 = Power_c_comp_4_2/self.Efficiency_comp #total power in kW
                if A_c_comp_1_2>=A_c_comp_4_2:
                    A_c_comp=A_c_comp_1_2
                else:
                    A_c_comp=A_c_comp_4_2
                    
                # print ('work of compressor is', Work_c_comp,'J/kg')
                # print ('Adjusted storage capacity is', H2_c_Cap_Storage, 'kg')
                # print ('flow rate is', m_c_flow_rate_1_2, 'and', m_c_flow_rate_4_2, 'kg/s')
                # print('Total fluid power of compressor', A_c_comp, 'kW')
                Number_c_Compressors=np.floor(A_c_comp/3000) #Number of compressors excluding the last one
                A_c_comp_1 = A_c_comp%3000  #power of the last compressor
                # print('Number of compressors', Number_c_Compressors+1)
                k1=2.2897
                k2=1.3604
                k3=-0.1027
                Compr_c_Cap_Cost=(10**(k1+k2*np.log10(3000)+k3*(np.log10(3000))**2))*Number_c_Compressors
                Compr_c_Cap_Cost_1=(10**(k1+k2*np.log10(A_c_comp_1)+k3*(np.log10(A_c_comp_1))**2))

                compressor_energy_used_1 = Work_c_comp*H2_c_Cap_Storage*2.8e-7
                compressor_energy_used_2 = Work_c_comp*H2_c_Cap_Storage*Release_efficiency*2.8e-7

                Compr_c_Energy_Costs_1 = compressor_energy_used_1*Energy_cost  #compressor electricity cost in cycle 1
                Compr_c_Energy_Costs_2 = compressor_energy_used_2*Energy_cost #compressor electricity cost assuming in regular charging cycle 
                
                Total_c_Compr_Cap_Cost = Compr_c_Cap_Cost + Compr_c_Cap_Cost_1
                Total_c_Compr_Cap_Cost = Total_c_Compr_Cap_Cost*(self.CEPCI_current/self.CEPCI2001)  ##Inflation
            else:
                Power_c_comp_1_2=0 #mechanical power of the pump in kW
                Power_c_comp_4_2=0
                A_c_comp_1_2 = 0  #total power in kW
                A_c_comp_4_2 = 0 #total power in kW
                Work_c_comp = 0
                Compr_c_Cap_Cost = 0
                compressor_energy_used_1 = 0
                compressor_energy_used_2 = 0
                Compr_c_Energy_Costs_1 = 0
                Compr_c_Energy_Costs_2 = 0
                Total_c_Compr_Cap_Cost = 0

            self.total_compressor_energy_used_kwh = compressor_energy_used_1 #+ compressor_energy_used_2

            # print ('Compressor energy cost is $', Compr_c_Energy_Costs)
            # print ('refrigeration capcost for compressor is $')
            # print('compressor capcost is $', Total_c_Compr_Cap_Cost)
            # print("----------")
            
            ########################################Costs associated with storage tanks
            
            # print("Number of tanks is: ", number_c_of_tanks)
            Storage_c_Tank_Cap_Costs=Cost_c_tank*number_c_of_tanks*self.Tank_manufacturing
            Storage_c_Tank_Cap_Costs = Storage_c_Tank_Cap_Costs*(self.CEPCI_current/self.CEPCI2007)  ##Inflation
            # print('Capcost for storage tank is', Storage_c_Tank_Cap_Costs)
            # print("----------")
            
            ###############################Refrigeration costs estimation adsorption process     
            # print ('pre-conditioning time is', round (t_precondition_hr), 'hr')
            H2_Cap=(capacity)    
            Ref_c_P_net_1_2=-(deltaP_c_net_1_2-Power_c_comp_1_2)  #Refrigeration power in kW from state 1 to state 2 (precondition)
            Ref_c_P_net_4_2=-(deltaP_c_net_4_2-Power_c_comp_4_2)  #Refrigeration power in kW from state 1 to state 2 (normal charging)
            if Ref_c_P_net_1_2 >= Ref_c_P_net_4_2:    
                Net_c_Cooling_Power_Adsorption=Ref_c_P_net_1_2 #Net refrigeration power in kW
            else:
                Net_c_Cooling_Power_Adsorption=Ref_c_P_net_4_2
            # print ('Net Cooling power for refrigeration sizing is',Net_c_Cooling_Power_Adsorption,'kW') #Cooling power in kW
            
            if Net_c_Cooling_Power_Adsorption < 1000:
                A1=-3.53E-09
                A2=-9.94E-06
                A3=3.30E-03
                nc=(A1*(self.Temp_c**3))+(A2*(self.Temp_c**2))+A3*self.Temp_c #Carnot efficiency factor
                COP=(self.Temp_c/(318-self.Temp_c))*nc #Coefficient of performance
                B1=24000
                B2=3500
                B3=0.9
                Total_c_Refrig_Cap_Costs_adsorption=(B1+(B2*(Net_c_Cooling_Power_Adsorption/COP)**B3))*(self.CEPCI_current/550.8)   
            else:
                Total_c_Refrig_Cap_Costs_adsorption=2*10**11*self.Temp_c**-2.077*(Net_c_Cooling_Power_Adsorption/1000)**0.6
                Total_c_Refrig_Cap_Costs_adsorption = Total_c_Refrig_Cap_Costs_adsorption*(self.CEPCI_current/self.CEPCI2017) 
            
            ####Utility for refrigeration
            Utility_c_ref = 4.07*10**7*self.Temp_c**(-2.669) #Utility in $/GJ, here, the utility is mostly for energy assumes 16.8 $/GJ (57 $/MWh)
            # Utility_c_refrigeration_1 = (self.CEPCI_current/self.CEPCI2017)*Utility_c_ref*-(deltaE_c_net_1_2-Work_c_comp*H2_c_Cap_Storage/1000)/1e6  
            energy_consumption_refrigeration_1_kj = -(deltaE_c_net_1_2-Work_c_comp*H2_c_Cap_Storage/1000) # in kJ
            
            Utility_c_refrigeration_1 = (Energy_cost/0.057)*Utility_c_ref*energy_consumption_refrigeration_1_kj/1e6  # changed based on discussion with original author 20221216, energy separated out 20230317
            # print ('refrigerator capital cost for adsorption is $', Total_c_Refrig_Cap_Costs_adsorption)    
            # print("------------")
            
            # Utility_c_refrigeration_2 = (self.CEPCI_current/self.CEPCI2017)*Utility_c_ref*-(deltaE_c_net_4_2-Work_c_comp*H2_c_Cap_Storage*Release_efficiency/1000)/1e6 
            energy_consumption_refrigeration_2_kj = -(deltaE_c_net_4_2-Work_c_comp*H2_c_Cap_Storage*Release_efficiency/1000) # in kJ
            Utility_c_refrigeration_2 = (Energy_cost/0.057)*Utility_c_ref*energy_consumption_refrigeration_2_kj/1e6  # changed based on discussion with original author 20221216, energy separated out 20230317
            
            # specify energy usage separately so energy usage can be used externally if desired
            joule2watthour = 1.0/3600.0 # 3600 joules in a watt hour (as also 3600 kJ in a kWh)
            energy_consumption_refrigeration_1_kwh = energy_consumption_refrigeration_1_kj*joule2watthour
            energy_consumption_refrigeration_2_kwh = energy_consumption_refrigeration_2_kj*joule2watthour
            self.total_refrigeration_energy_used_kwh = energy_consumption_refrigeration_1_kwh #+ energy_consumption_refrigeration_2_kwh
            
            if self.total_refrigeration_energy_used_kwh < 0:
                raise(ValueError("energy usage must be greater than 0"))
            ###############################Heating costs desorption process   
            k1=6.9617
            k2=-1.48
            k3=0.3161
            Net_c_Heating_Power_Desorption=detlaP_c_net_2_3/self.Efficiency_heater ## steam boiler power at 0.7 efficiency in kW
            Number_c_Heaters=np.floor(Net_c_Heating_Power_Desorption/9400) #Number of compressors excluding the last one
            Heater_c_Power_1 = Net_c_Heating_Power_Desorption%9400  #power of the last compressor
            # print('Number of heaters', Number_c_Heaters+1)
            Heater_c_Cap_Cost=(10**(k1+k2*np.log10(9400)+k3*(np.log10(9400))**2))*Number_c_Heaters
            if Heater_c_Power_1 < 1000:
                Heater_c_Cap_Cost_1=(10**(k1+k2*np.log10(1000)+k3*(np.log10(1000))**2))*(Heater_c_Power_1/1000)
            else:
                Heater_c_Cap_Cost_1=(10**(k1+k2*np.log10(Heater_c_Power_1)+k3*(np.log10(Heater_c_Power_1))**2))
            Total_c_Heater_Cap_Cost = Heater_c_Cap_Cost + Heater_c_Cap_Cost_1
            Total_c_Heater_Cap_Cost = Total_c_Heater_Cap_Cost *(self.CEPCI_current/self.CEPCI2001)  ##Inflation #TODO make inflation optional per user input
            
            Utility_c_Heater = 0 # Jared Thomas set to zero as per discussion with Peng Peng through Abhineet Gupta 20221215 was 13.28*deltaE_c_net_2_3/1e6 #$13.28/GJ for low pressure steam
            self.total_heating_energy_used_kwh = Net_c_Heating_Power_Desorption*t_discharge_hr
            Total_c_Heating_Energy_Costs = self.total_heating_energy_used_kwh*Energy_cost
            
            # print('heater capcost is $', Total_c_Heater_Cap_Cost)
            
            ########################################Operational costs (sized based on cycle 1 requirements)###########################################
            Op_c_Costs_1 = Compr_c_Energy_Costs_1 + Utility_c_refrigeration_1 + Utility_c_Heater + Total_c_Heating_Energy_Costs
            Op_c_Costs_2 = Compr_c_Energy_Costs_2 + Utility_c_refrigeration_2 + Utility_c_Heater + Total_c_Heating_Energy_Costs
            Total_c_Cap_Costs = Storage_c_Tank_Cap_Costs + Total_c_Refrig_Cap_Costs_adsorption + Total_c_Compr_Cap_Cost + Total_c_Heater_Cap_Cost
            
            # Op_c_Costs = (Op_c_Costs_1 + Op_c_Costs_2 * (cycle_number-1)+self.maintanance*Total_c_Cap_Costs+self.wage*360*2)/cycle_number/capacity
            #TODO check this. I changed the 2 to a 24 because it looks like it should be working hours in a year.
            Op_c_Costs = ((Op_c_Costs_1 + Op_c_Costs_2*(cycle_number-1) + self.maintanance*Total_c_Cap_Costs + self.wage*360*2)/cycle_number) # checked, this was divided by capacity, but Peng Peng confirmed it was duplicating the following divisions by capacity
            
            ######################writing costs#####################################################
            self.cost_kg[i] = (Total_c_Cap_Costs/capacity + self.Site_preparation)*self.Markup
            cost_kg_tank[i] = Storage_c_Tank_Cap_Costs/capacity
            cost_kg_comp[i] = Total_c_Compr_Cap_Cost/capacity
            cost_kg_ref[i] = Total_c_Refrig_Cap_Costs_adsorption/capacity    
            cost_kg_heat[i] = Total_c_Heater_Cap_Cost/capacity
            self.Op_c_Costs_kg[i] = Op_c_Costs/capacity
            # print("\n Pressure Vessel Costs: ")
            # print("cost_kg ")
            # print("cost_kg_tank ")
            # print("cost_kg_comp ")
            # print("cost_kg_ref ")
            # print("cost_kg_heat ")
            ######################################## Total Energy Use (kWh) ######################
            self.total_energy_used_kwh[i] = self.total_compressor_energy_used_kwh + self.total_heating_energy_used_kwh + self.total_refrigeration_energy_used_kwh
        
        self.curve_fit()

    def curve_fit(self):

        ################### plot prep ###########
        self.plot_range = range(int(np.min(self.capacity_1)), int(np.max(self.capacity_1)), 100)

        ###################Fitting capital####################################################  

        var_cap = [0.01 ,0.5, 5]   #Initial guesses for the parameters, can be flexible
        
        varfinal_cap_fitted, success = leastsq(self.residual_op, var_cap, args=(self.capacity_1, self.cost_kg), maxfev=100000)
        
        self.a_cap_fit=varfinal_cap_fitted[0]
        self.b_cap_fit=varfinal_cap_fitted[1]
        self.c_cap_fit=varfinal_cap_fitted[2] 

        if self.verbose:
            print ('a_cap is', self.a_cap_fit)
            print ('b_cap is', self.b_cap_fit)
            print ('c_cap is', self.c_cap_fit)
            print ('***********')

        self.fitted_capex = self.exp_log_fit(varfinal_cap_fitted, self.plot_range)      
        
        # popt, pcov = curve_fit(self.exp_fit, self.capacity_1, self.cost_kg, maxfev=100000)
        
        # self.a_cap_fit=popt[0]
        # self.b_cap_fit=popt[1]
        
        # print ('a is', self.a_cap_fit)
        # print ('b is', self.b_cap_fit)
        # print ('***********')
        
        # self.fitted_kg = self.exp_fit(self.plot_range,self.a_cap_fit,self.b_cap_fit)
        
        ####################### fitting OpEx #################################
        var_op= [0.01 ,0.5, 5]   #Initial guesses for the parameters, can be flexible
        
        varfinal_op_fitted,success = leastsq(self.residual_op, var_op, args=(self.capacity_1, self.Op_c_Costs_kg), maxfev=100000)
        
        self.a_op_fit=varfinal_op_fitted[0]
        self.b_op_fit=varfinal_op_fitted[1]
        self.c_op_fit=varfinal_op_fitted[2] 

        if self.verbose:
            print ('a_op is', self.a_op_fit)
            print ('b_op is', self.b_op_fit)
            print ('c_op is', self.c_op_fit)
            print ('***********')

        self.fitted_op_kg = self.exp_log_fit(varfinal_op_fitted, self.plot_range)
  
        ##################### Fit energy usage ################################
        self.energy_coefficients = np.polyfit(self.capacity_1, self.total_energy_used_kwh, 1)
        self.energy_function = np.poly1d(self.energy_coefficients) # kWh
        self.fit_energy_wrt_capacity_kwh = self.energy_function(self.plot_range)
    
    def plot(self):

        fig, ax = plt.subplots(2,2, sharex=True, figsize=(10,6))
        
        ##################### CAPEX #######################
        ax[0,0].scatter(self.capacity_1*1E-3, self.cost_kg, color='r', label = 'Calc')
        ax[0,0].plot(np.asarray(self.plot_range)*1E-3, self.fitted_capex, label = 'Fit')
        # ax[0,0].plot(self.capacity_1,cost_kg_tank, color='b', label = 'tank')
        # ax[0,0].plot(self.capacity_1,cost_kg_comp, color='c', label = 'compressor')
        # ax[0,0].plot(self.capacity_1,cost_kg_ref, color='m', label = 'refrigeration')
        # ax[0,0].plot(self.capacity_1,cost_kg_heat, color='y', label = 'heater')   
        
        a_disp=np.round(self.a_cap_fit, 2)
        b_disp=np.round(self.b_cap_fit, 2)
        # plt.ylim(0,np.amax(self.cost_kg)*2)
        # equation_cap = 'y='+str(a_disp)+'x'+'^'+str(b_disp)
        a_cap_fit_disp=np.round(self.a_cap_fit, 2)
        b_cap_fit_disp=np.round(self.b_cap_fit, 2)
        c_cap_fit_disp=np.round(self.c_cap_fit, 2) 
        equation_cap = 'y='+'exp('+str(a_cap_fit_disp)+'(ln(x))^2\n-'+str(b_cap_fit_disp)+'ln(x)+' + str(c_cap_fit_disp) +')'

        ax[0,0].annotate(equation_cap, xy=(np.amax(self.capacity_1)*1E-3*0.4 , np.amax(self.cost_kg)*0.8)) 
        
        ax[0,0].set_ylabel("CAPEX ($/kg)")
        ax[0,0].legend(loc='best', frameon=False)
        # plt.legend(loc='best')
        # ax[0,0].title('Capital')

        ##################### OPEX ############################
        a_op_fit_disp=np.round(self.a_op_fit, 2)
        b_op_fit_disp=np.round(self.b_op_fit, 2)
        c_op_fit_disp=np.round(self.c_op_fit, 2) 
        
        equation_op = 'y='+'exp('+str(a_op_fit_disp)+'(ln(x))^2\n-'+str(b_op_fit_disp)+'ln(x)+' + str(c_op_fit_disp) +')'
        
        ax[0,1].plot(np.asarray(self.plot_range)*1E-3, self.fitted_op_kg, label = 'Fit')
        ax[0,1].scatter(self.capacity_1*1E-3, self.Op_c_Costs_kg, color='r', label = 'Calc')     
        ax[0,1].set_ylabel("OPEX ($/kg)")
        ax[0,1].annotate(equation_op, xy=(np.amax(self.capacity_1)*1E-3*0.2 , np.amax(self.Op_c_Costs_kg)*0.4))     
        ax[0,1].legend(loc='best', frameon=False)
        # plt.legend(loc='best')
        # ax[0,1].title('Annual operational')

        ################## Energy ######################  
        ax[1,1].plot(np.asarray(self.plot_range)*1E-3, self.fit_energy_wrt_capacity_kwh*1E-6, label = 'Fit')
        ax[1,1].scatter(self.capacity_1*1E-3, self.total_energy_used_kwh*1E-6, color='r', label = 'Calc')    
        ax[1,1].set_xlabel("Capacity (Tonnes H2)")    
        ax[1,1].set_ylabel("Energy Use (GWh)")
        
        equation_energy = 'y='+str(round(self.energy_coefficients[0],2)) +'x+'+str(round(self.energy_coefficients[1],2)) 
        ax[1,1].annotate(equation_energy, xy=(3000, 5))     
        ax[1,1].legend(loc='best', frameon=False)
        ax[1,1].legend(loc='best', frameon=False)
        # ax[1,1].title('Annual operational')

        ################ Wrap Up ######################
        
        ax[1,0].set_xlabel("Capacity (Tonnes H2)")    

        plt.tight_layout()
        plt.show()



from greenheart.simulation.technologies.hydrogen.h2_storage.pressure_vessel.compressed_gas_storage_model_20221021.Compressed_all import PressureVessel
from greenheart.simulation.technologies.hydrogen.h2_storage.pressure_vessel.compressed_gas_storage_model_20221021.Compressed_gas_function import CompressedGasFunction



"""
Function to call profast for ammonia model
Written by Abhineet Gupta
"""

import ProFAST
import pandas as pd

# # Add location of PyFAST code 
# import sys
# sys.path.append('../PyFAST/')

# import src.PyFAST as PyFAST

# Implement equations from Ammonia model received
def run_profast_for_ammonia(plant_capacity_kgpy,plant_capacity_factor,plant_life,levelized_cost_of_hydrogen, electricity_cost,grid_prices_interpolated_USDperMWh,cooling_water_cost,iron_based_catalyst_cost,oxygen_price):
    # Inputs:
    # plant_capacity_kgpy = 462323016 ##KgNH3/year
    # plant_capacity_factor = 0.9
    # plant_life = 40 #years
    # # Costs from original model (for reference)
    # levelized_cost_of_hydrogen = 4.83       # $/kg
    # electricity_cost = 69.83                # $/MWh
    # cooling_water_cost = 0.000113349938601175 # $/Gal
    # iron_based_catalyist_cost = 23.19977341 # $/kg
    # oxygen_price = 0.0285210891617726       # $/kg

    model_year_CEPCI = 596.2
    equation_year_CEPCI = 541.7
    
    ammonia_production_kgpy = plant_capacity_kgpy*plant_capacity_factor #=  416,090,714      
    
    # scale with respect to a baseline plant
    scaling_ratio = plant_capacity_kgpy/(365.0*1266638.4)
    
    # CapEx
    scaling_factor_equipment = 0.6
    capex_scale_factor = scaling_ratio**scaling_factor_equipment
    capex_air_separation_crygenic = model_year_CEPCI/equation_year_CEPCI*22506100 * capex_scale_factor
    capex_haber_bosch = model_year_CEPCI/equation_year_CEPCI*18642800 * capex_scale_factor
    capex_boiler = model_year_CEPCI/equation_year_CEPCI*7069100 * capex_scale_factor
    capex_cooling_tower = model_year_CEPCI/equation_year_CEPCI*4799200 * capex_scale_factor
    capex_direct = capex_air_separation_crygenic + capex_haber_bosch\
        + capex_boiler + capex_cooling_tower
    capex_depreciable_nonequipment = capex_direct*0.42 + \
       4112701.84103543*scaling_ratio
    capex_total = capex_direct + capex_depreciable_nonequipment
    
    land_cost = capex_depreciable_nonequipment
    
    # O&M Cost
    scaling_factor_labor = 0.25
    labor_cost = 57 * 50 * 2080 * scaling_ratio**scaling_factor_labor
    general_administration_cost = labor_cost * 0.2
    property_tax_insurance = capex_total * 0.02
    maintenance_cost = capex_direct * 0.005 * \
        scaling_ratio**scaling_factor_equipment
    land_cost = 2500000*capex_scale_factor
    fixed_O_and_M_cost = land_cost + labor_cost + \
        general_administration_cost + \
        property_tax_insurance + maintenance_cost
    
    # Feedstock
    H2_consumption = 0.197284403 # kg_H2/ kg_NH3
    H2_cost_in_startup_year = levelized_cost_of_hydrogen * H2_consumption\
         * plant_capacity_kgpy * plant_capacity_factor
    
    electricity_usage = 0.530645243/1000 # mWh/kg_NH3
    energy_cost_in_startup_year = electricity_cost * electricity_usage\
        * plant_capacity_kgpy * plant_capacity_factor # 
    
    cooling_water_usage = 0.049236824 # Gal/kg_NH3
    iron_based_catalyst_usage = 0.000091295354067341 # kg/kg_NH3
    non_energy_cost_in_startup_year = \
        ((cooling_water_cost * cooling_water_usage) + \
            (iron_based_catalyst_cost*iron_based_catalyst_usage)) * \
                plant_capacity_kgpy * plant_capacity_factor
    
    variable_cost_in_startup_year = energy_cost_in_startup_year\
        + non_energy_cost_in_startup_year

    # By-product
    oxygen_byproduct = 0.29405077250145     # kg/kg_NH#
    credits_byproduct = oxygen_price*oxygen_byproduct * \
        plant_capacity_kgpy * plant_capacity_factor

    financial_assumptions = pd.read_csv('H2_Analysis/financial_inputs.csv',index_col=None,header=0)
    financial_assumptions.set_index(["Parameter"], inplace = True)
    financial_assumptions = financial_assumptions['Hydrogen/Steel/Ammonia']

     # Set up ProFAST
    pf = ProFAST.ProFAST()

    install_years = 3
    analysis_start = list(grid_prices_interpolated_USDperMWh.keys())[0] - install_years
    
    # Fill these in - can have most of them as 0 also
    gen_inflation = 0.00
    pf.set_params('commodity',{"name":'Ammonia',"unit":"kg","initial price":1000,"escalation":gen_inflation})
    pf.set_params('capacity',plant_capacity_kgpy/365) #units/day
    pf.set_params('maintenance',{"value":0,"escalation":gen_inflation})
    pf.set_params('analysis start year',analysis_start)
    pf.set_params('operating life',plant_life)
    pf.set_params('installation months',12*install_years)
    pf.set_params('installation cost',{"value":fixed_O_and_M_cost,"depr type":"Straight line","depr period":4,"depreciable":False})
    pf.set_params('non depr assets',land_cost)
    pf.set_params('end of proj sale non depr assets',land_cost*(1+gen_inflation)**plant_life)
    pf.set_params('demand rampup',0)
    pf.set_params('long term utilization',plant_capacity_factor)
    pf.set_params('credit card fees',0)
    pf.set_params('sales tax',0) 
    pf.set_params('license and permit',{'value':00,'escalation':gen_inflation})
    pf.set_params('rent',{'value':0,'escalation':gen_inflation})
    pf.set_params('property tax and insurance',0)
    pf.set_params('admin expense',0)
    pf.set_params('total income tax rate',financial_assumptions['total income tax rate'])
    pf.set_params('capital gains tax rate',financial_assumptions['capital gains tax rate'])
    pf.set_params('sell undepreciated cap',True)
    pf.set_params('tax losses monetized',True)
    pf.set_params('general inflation rate',gen_inflation)
    pf.set_params('leverage after tax nominal discount rate',financial_assumptions['leverage after tax nominal discount rate'])
    pf.set_params('debt equity ratio of initial financing',financial_assumptions['debt equity ratio of initial financing'])
    pf.set_params('debt type','Revolving debt')
    pf.set_params('debt interest rate',financial_assumptions['debt interest rate'])
    pf.set_params('cash onhand',1)
    
    #----------------------------------- Add capital items to ProFAST ----------------
    pf.add_capital_item(name="Air Separation by Cryogenic",cost=capex_air_separation_crygenic,depr_type="MACRS",depr_period=7,refurb=[0])
    pf.add_capital_item(name="Haber Bosch",cost=capex_haber_bosch,depr_type="MACRS",depr_period=7,refurb=[0])
    pf.add_capital_item(name="Boiler and Steam Turbine",cost=capex_boiler,depr_type="MACRS",depr_period=7,refurb=[0])
    pf.add_capital_item(name="Cooling Tower",cost=capex_cooling_tower,depr_type="MACRS",depr_period=7,refurb=[0])
    pf.add_capital_item(name="Depreciable Nonequipment",cost=capex_depreciable_nonequipment,depr_type="MACRS",depr_period=7,refurb=[0])

    total_capex = capex_air_separation_crygenic+capex_haber_bosch+capex_boiler+capex_cooling_tower+capex_depreciable_nonequipment
    
    #-------------------------------------- Add fixed costs--------------------------------
    pf.add_fixed_cost(name="Labor Cost",usage=1,unit='$/year',cost=labor_cost,escalation=gen_inflation)
    pf.add_fixed_cost(name="Maintenance Cost",usage=1,unit='$/year',cost=maintenance_cost,escalation=gen_inflation)
    pf.add_fixed_cost(name="Administrative Expense",usage=1,unit='$/year',cost=general_administration_cost,escalation=gen_inflation)
    pf.add_fixed_cost(name="Property tax and insurance",usage=1,unit='$/year',cost=property_tax_insurance,escalation=0.0)
    #pf.add_fixed_cost(name="Land cost",cost=2500000*capex_scale_factor,depr_type="MACRS",depr_period=20,refurb=[0])
     
    # Putting property tax and insurance here to zero out depcreciation/escalation. Could instead put it in set_params if
    # we think that is more accurate
    
    #---------------------- Add feedstocks, note the various cost options-------------------
    pf.add_feedstock(name='Hydrogen',usage=H2_consumption,unit='kilogram of hydrogen per kilogram of ammonia',cost=levelized_cost_of_hydrogen,escalation=gen_inflation)
    pf.add_feedstock(name='Electricity',usage=electricity_usage,unit='MWh per kilogram of ammonia',cost=grid_prices_interpolated_USDperMWh,escalation=gen_inflation)
    pf.add_feedstock(name='Cooling water',usage=cooling_water_usage,unit='Gallon per kilogram of ammonia',cost=cooling_water_cost,escalation=gen_inflation)
    pf.add_feedstock(name='Iron based catalyst',usage=iron_based_catalyst_usage,unit='kilogram of catalyst per kilogram of ammonia',cost=iron_based_catalyst_cost,escalation=gen_inflation)
    pf.add_coproduct(name='Oxygen byproduct',usage=oxygen_byproduct,unit='kilogram of oxygen per kilogram of ammonia',cost=oxygen_price,escalation=gen_inflation)
    
    #------------------------------ Sovle for breakeven price ---------------------------
    
    sol = pf.solve_price()
    
    summary = pf.summary_vals
    
    price_breakdown = pf.get_cost_breakdown()
    
    price_breakdown_air_separation_by_cryogenic = price_breakdown.loc[price_breakdown['Name']=='Air Separation by Cryogenic','NPV'].tolist()[0]
    price_breakdown_Haber_Bosch = price_breakdown.loc[price_breakdown['Name']=='Haber Bosch','NPV'].tolist()[0]
    price_breakdown_boiler_and_steam_turbine = price_breakdown.loc[price_breakdown['Name']=='Boiler and Steam Turbine','NPV'].tolist()[0]
    price_breakdown_cooling_tower = price_breakdown.loc[price_breakdown['Name']=='Cooling Tower','NPV'].tolist()[0]
    price_breakdown_depreciable_nonequipment = price_breakdown.loc[price_breakdown['Name']=='Depreciable Nonequipment','NPV'].tolist()[0]
    price_breakdown_installation = price_breakdown.loc[price_breakdown['Name']=='Installation cost','NPV'].tolist()[0]
 

    price_breakdown_labor_cost_annual = price_breakdown.loc[price_breakdown['Name']=='Labor Cost','NPV'].tolist()[0]  
    price_breakdown_maintenance_cost = price_breakdown.loc[price_breakdown['Name']=='Maintenance Cost','NPV'].tolist()[0]  
    price_breakdown_administrative_expense = price_breakdown.loc[price_breakdown['Name']=='Administrative Expense','NPV'].tolist()[0]  
    price_breakdown_property_tax_and_insurance = price_breakdown.loc[price_breakdown['Name']=='Property tax and insurance','NPV'].tolist()[0]
    #price_breakdown_land_cost = price_breakdown.loc[price_breakdown['Name']=='Land cost','NPV'].tolist()[0]  

    if levelized_cost_of_hydrogen < 0:
        price_breakdown_hydrogen = -1*price_breakdown.loc[price_breakdown['Name']=='Hydrogen','NPV'].tolist()[0] 
    else:
        price_breakdown_hydrogen = price_breakdown.loc[price_breakdown['Name']=='Hydrogen','NPV'].tolist()[0] 
    price_breakdown_electricity = price_breakdown.loc[price_breakdown['Name']=='Electricity','NPV'].tolist()[0] 
    price_breakdown_cooling_water = price_breakdown.loc[price_breakdown['Name']=='Cooling water','NPV'].tolist()[0]
    price_breakdown_iron_based_catalyst = price_breakdown.loc[price_breakdown['Name']=='Iron based catalyst','NPV'].tolist()[0]
    price_breakdown_oxygen_byproduct = -price_breakdown.loc[price_breakdown['Name']=='Oxygen byproduct','NPV'].tolist()[0]

    price_breakdown_taxes = price_breakdown.loc[price_breakdown['Name']=='Income taxes payable','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name'] == 'Monetized tax losses','NPV'].tolist()[0]\
        
    if gen_inflation > 0:
        price_breakdown_taxes = price_breakdown_taxes + price_breakdown.loc[price_breakdown['Name']=='Capital gains taxes payable','NPV'].tolist()[0]
        
    # price_breakdown_financial = price_breakdown.loc[price_breakdown['Name']=='Non-depreciable assets','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Cash on hand reserve','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Property tax and insurance','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Repayment of debt','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Interest expense','NPV'].tolist()[0]\
    #     + price_breakdown.loc[price_breakdown['Name']=='Dividends paid','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Sale of non-depreciable assets','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Cash on hand recovery','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Inflow of debt','NPV'].tolist()[0]\
    #     - price_breakdown.loc[price_breakdown['Name']=='Inflow of equity','NPV'].tolist()[0]

        # Calculate financial expense associated with equipment
    price_breakdown_financial_equipment = price_breakdown.loc[price_breakdown['Name']=='Repayment of debt','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Interest expense','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Dividends paid','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Inflow of debt','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Inflow of equity','NPV'].tolist()[0]    
        
    # Calculate remaining financial expenses
    price_breakdown_financial_remaining = price_breakdown.loc[price_breakdown['Name']=='Non-depreciable assets','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Cash on hand reserve','NPV'].tolist()[0]\
        + price_breakdown.loc[price_breakdown['Name']=='Property tax and insurance','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Sale of non-depreciable assets','NPV'].tolist()[0]\
        - price_breakdown.loc[price_breakdown['Name']=='Cash on hand recovery','NPV'].tolist()[0]
        
    price_check = price_breakdown_air_separation_by_cryogenic+price_breakdown_Haber_Bosch+price_breakdown_boiler_and_steam_turbine+price_breakdown_cooling_tower+price_breakdown_depreciable_nonequipment\
        +price_breakdown_installation+price_breakdown_labor_cost_annual+price_breakdown_maintenance_cost+price_breakdown_administrative_expense\
        +price_breakdown_hydrogen+price_breakdown_electricity+price_breakdown_cooling_water+price_breakdown_iron_based_catalyst-price_breakdown_oxygen_byproduct\
        +price_breakdown_taxes+price_breakdown_financial_equipment+price_breakdown_financial_remaining
        
    ammonia_price_breakdown = {'Ammonia price: Air Separation by Cryogenic ($/kg)':price_breakdown_air_separation_by_cryogenic,
                               'Ammonia price: Haber Bosch ($/kg)':price_breakdown_Haber_Bosch,'Ammonia price: Boiler and Steam Turbine ($/kg)':price_breakdown_boiler_and_steam_turbine,
                               'Ammonia price: Cooling Tower ($/kg)':price_breakdown_cooling_tower,'Ammonia price: Depreciable Nonequipment ($/kg)':price_breakdown_depreciable_nonequipment,
                               'Ammonia price: Labor Cost ($/kg)':price_breakdown_labor_cost_annual,'Ammonia price: Maintenance Cost ($/kg)':price_breakdown_maintenance_cost,
                               'Ammonia price: Administrative Expense ($/kg)':price_breakdown_administrative_expense,
                               'Ammonia price: Hydrogen ($/kg)':price_breakdown_hydrogen,'Ammonia price: Electricity ($/kg)':price_breakdown_electricity,
                               'Ammonia price: Cooling water ($/kg)':price_breakdown_cooling_water,'Ammonia price: Iron based catalyst ($/kg)':price_breakdown_iron_based_catalyst,
                               'Ammonia price: Oxygen byproduct ($/kg)':price_breakdown_oxygen_byproduct,'Ammonia price: Taxes ($/kg)':price_breakdown_taxes,
                               'Ammonia price: Equipment Financing ($/kg)':price_breakdown_financial_equipment,
                               'Ammonia price: Remaining Financial ($/kg)':price_breakdown_financial_remaining,'Ammonia price: Total ($/kg)':price_check}
    
    price_breakdown = price_breakdown.drop(columns=['index','Amount'])
    
    return(sol,summary,price_breakdown,ammonia_production_kgpy,ammonia_price_breakdown,total_capex)


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 17 20:08:09 2022
@author: ppeng

Revisions:
- 20221118:
    Author: Jared J. Thomas
    Description: 
        - Reformatted to be a class
"""

"""
Model Revision Needed: storage space and mass
Description: This file should handle physical size (footprint and mass) needed for pressure vessel storage
Sources:
    - [1] ./README.md and other elements in this directory
Args:
    - same as for the physics and cost model contained herein
    - others may be added as needed
Returns:(can be from separate functions and/or methods as it makes sense):
    - mass_empty (float): mass (approximate) for pressure vessel storage components ignoring stored H2
    - footprint (float): area required for pressure vessel storage
    - others may be added as needed
"""

# package imports
import os
import numpy as np

# local imports
# from Compressed_gas_function import CompressedGasFunction
from .Compressed_gas_function import CompressedGasFunction

class PressureVessel():
    def __init__(self, Wind_avai=80, H2_flow=200, cdratio=1, Energy_cost=0.07, cycle_number=1, parent_path=os.path.abspath(os.path.dirname(__file__)), spread_sheet_name="Tankinator.xlsx", verbose=False):

        ########Key inputs##########
        self.Wind_avai = Wind_avai  #Wind availability in %
        self.H2_flow = H2_flow  #Flow rate of steel plants in tonne/day
        self.cdratio = cdratio  #Charge/discharge ratio, for example 2 means the charging is 2x faster than discharge
        self.Energy_cost = Energy_cost  #Renewable energy cost in $/kWh

        #######Other inputs########
        self.cycle_number = cycle_number #Equivalent cycle number for a year, only affects operation (the higher the number is the less effect there will be), set as now as I am not sure how the maximum sotrage capacity is determined and how the storage will be cycled

        self.compressed_gas_function = CompressedGasFunction(path_tankinator=os.path.join(parent_path, spread_sheet_name))
        self.compressed_gas_function.verbose = verbose
        
    def run(self):
        #####Run calculation########
        self.compressed_gas_function.func(Wind_avai=self.Wind_avai, H2_flow=self.H2_flow, cdratio=self.cdratio, Energy_cost=self.Energy_cost, cycle_number=self.cycle_number)

        ########Outputs################

        ######Maximum equivalent storage capacity and duration
        self.capacity_max = self.compressed_gas_function.capacity_max   #This is the maximum equivalent H2 storage in kg
        self.t_discharge_hr_max = self.compressed_gas_function.t_discharge_hr_max   #This is tha maximum storage duration in kg

        ###Parameters for capital cost fitting for optimizing capital cost
        self.a_fit_capex = self.compressed_gas_function.a_cap_fit
        self.b_fit_capex = self.compressed_gas_function.b_cap_fit
        self.c_fit_capex = self.compressed_gas_function.c_cap_fit

        #Parameters for operational cost fitting for optimizing capital cost
        self.a_fit_opex = self.compressed_gas_function.a_op_fit
        self.b_fit_opex = self.compressed_gas_function.b_op_fit
        self.c_fit_opex = self.compressed_gas_function.c_op_fit

    def calculate_from_fit(self, capacity_kg):
        capex_per_kg = self.compressed_gas_function.exp_log_fit([self.a_fit_capex, self.b_fit_capex, self.c_fit_capex], capacity_kg) 
        opex_per_kg = self.compressed_gas_function.exp_log_fit([self.a_fit_opex, self.b_fit_opex, self.c_fit_opex], capacity_kg) 
        energy_per_kg_h2 = self.compressed_gas_function.energy_function(capacity_kg)/capacity_kg

        # NOTE ON ENERGY: the energy value returned here is the energy used to fill the 
        # tanks initially for the first fill and so can be used as an approximation for the energy used on a per kg basis. 
        # If cycle_number > 1, the energy model output is incorrect.

        capex = capex_per_kg*capacity_kg
        opex = opex_per_kg*capacity_kg
        return capex, opex, energy_per_kg_h2

    def get_tanks(self, capacity_kg):
        """ gets the number of tanks necessary """
        return np.ceil(capacity_kg/self.compressed_gas_function.m_H2_tank)

    def get_tank_footprint(self, capacity_kg,
                           upright : bool = True,
                           custom_packing : bool = False,
                           packing_ratio : float = None):
        """
        gets the footprint required for the H2 tanks

        assumes that packing is square (unless custom_packing is true)
        - diameter D upright tank occupies D^2
        - diameter D, length L tank occupies D*L

        parameters:
            - `upright`: place tanks vertically (default yes)?
            - `custom_packing`: pack tanks at an alternate packing fraction?
            - `packing_ratio`: ratio for custom packing, defaults to theoretical max (if known)
        returns:
            - `tank_footprint`: footprint of each tank in m^2
            - `array_footprint`: total footprint of all tanks in m^2
        """

        tank_radius= self.compressed_gas_function.Router/100
        tank_length= self.compressed_gas_function.Louter/100
        Ntank= self.get_tanks(capacity_kg= capacity_kg)

        if upright:
            tank_area= np.pi*tank_radius**2
            tank_footprint= 4*tank_radius**2
        else:
            tank_area= np.pi*tank_radius**2*((tank_length - 2*tank_radius)*(2*tank_radius))
            tank_footprint= tank_radius*tank_length

        if custom_packing:
            if upright:
                if packing_ratio is None: packing_ratio= np.pi*np.sqrt(3.)/6. # default to tight packing
                tank_footprint= tank_area*packing_ratio
            else:
                if packing_ratio is None:
                    raise NotImplementedError("tight packing ratio for cylinders isn't derived yet")
                tank_footprint= tank_area*packing_ratio

        return (tank_footprint, Ntank*tank_footprint)
    
    def get_tank_mass(self, capacity_kg):
        """
        gets the mass required for the H2 tanks

        returns
            - `tank_mass`: mass of each tank
            - `array_mass`: total mass of all tanks
        """

        tank_mass = self.compressed_gas_function.Mempty_tank
        Ntank = self.get_tanks(capacity_kg = capacity_kg)

        return (tank_mass, Ntank*tank_mass)

    def plot(self):
        self.compressed_gas_function.plot()

    def distributed_storage_vessels(self, capacity_total_tgt, N_sites):
        """
        compute modified pressure vessel storage requirements for distributed
        pressure vessels

        parameters:
            - capacity_total_tgt: target gaseous H2 capacity in kilograms
            - N_sites: number of sites (e.g. turbines) where pressure vessels will be placed

        returns:
            - 
        """

        # assume that the total target capacity is equally distributed across sites
        capacity_site_tgt= capacity_total_tgt/N_sites

        # capex_centralized_total, opex_centralized_total, energy_kg_centralized_total= self.calculate_from_fit(capacity_total_tgt)
        capex_site, opex_site, energy_kg_site= self.calculate_from_fit(capacity_site_tgt)

        # get the resulting capex & opex costs, incl. equivalent
        capex_distributed_total= N_sites*capex_site # the cost for the total distributed storage facilities
        opex_distributed_total= N_sites*opex_site # the cost for the total distributed storage facilities

        # get footprint stuff
        area_footprint_site= self.get_tank_footprint(capacity_site_tgt)[1]
        mass_tank_empty_site= self.get_tank_mass(capacity_site_tgt)[1]

        # return the outputs
        return capex_distributed_total, opex_distributed_total, energy_kg_site, \
                area_footprint_site, mass_tank_empty_site, capacity_site_tgt

if __name__ == "__main__":
    storage = PressureVessel()
    storage.run()

    capacity_req= 1e3
    print("tank type:", storage.compressed_gas_function.tank_type)
    print("tank mass:", storage.get_tank_mass(capacity_req)[0])
    print("tank radius:", storage.compressed_gas_function.Router)
    print("tank length:", storage.compressed_gas_function.Louter)
    print("tank footprint (upright):", storage.get_tank_footprint(capacity_req, upright= True)[0])
    print("tank footprint (flat):", storage.get_tank_footprint(capacity_req, upright= False)[0])
    
    print("\nnumber of tanks req'd:",
          storage.get_tanks(capacity_req))
    print("total footprint (upright):", storage.get_tank_footprint(capacity_req, upright= True)[1])
    print("total footprint (flat):", storage.get_tank_footprint(capacity_req, upright= False)[1])
    print("total mass:", storage.get_tank_mass(capacity_req)[1])
